

# Built-in
import os
import warnings


# Common
import numpy as np
import scipy.constants as scpct
from scipy.interpolate import BSpline
import matplotlib.pyplot as plt


# specific
from . import _fit12d_funccostjac as _funccostjac


__all__ = [
    'fit1d_extract',
    'fit2d_extract',
]


# Think this through again:
# automatically load all ?
# width => Ti?
# shift => vi?
_D3 = {
    'bck_amp': {
        'types': ['x'],
        'unit': 'a.u.',
        'field': 'bck_amp',
    },
    'bck_rate': {
        'types': ['x'],
        'unit': 'a.u.',
        'field': 'bck_rate',
    },
    'amp': {
        'types': ['x', 'lines'],
        'units': 'a.u.',
        'field': 'amp',
    },
    'width': {
        'types': ['x', 'lines'],
        'units': 'a.u.',
        'field': 'width',
    },
    'shift': {
        'types': ['x', 'lines'],
        'units': 'a.u.',
        'field': 'shift',
    },
    'ratio': {
        'types': ['lines'],
        'units': 'a.u.',
        'field': 'amp',
    },
    'Ti': {
        'types': ['lines'],
        'units': 'eV',
        'field': 'width',
    },
    'vi': {
        'types': ['x'],
        'units': 'm.s^-1',
        'field': 'shift',
    },
    'dratio': {
        'types': ['x'],
        'units': 'a.u.',
        'field': 'dratio',
    },
    'dshift': {
        'types': ['x'],
        'units': 'a.u.',
        'field': 'dshift',
    },
}
_ALLOW_PICKLE = True


###########################################################
###########################################################
#
#   Extract data from pre-computed dict of fitted results
#
###########################################################
###########################################################


def fit12d_get_data_checkformat(
    dfit=None,
    bck=None,
    amp=None, ratio=None,
    Ti=None, width=None,
    vi=None, shift=None,
    sol_total=None,
    sol_detail=None,
    sol_pts=None,
    phi_prof=None,
    phi_npts=None,
    vs_nbs=None,
    allow_pickle=None,
):

    # ----------------
    # load file if str

    if isinstance(dfit, str):
        if not os.path.isfile(dfit) or not dfit[-4:] == '.npz':
            msg = ("Provided dfit must be either a dict or "
                   + "the absolute path to a saved .npz\n"
                   + "  You provided: {}".format(dfit))
            raise Exception(msg)
        if allow_pickle is None:
            allow_pickle = _ALLOW_PICKLE
        dfit = dict(np.load(dfit, allow_pickle=allow_pickle))
        _rebuild_dict(dfit)

    # ----------------
    # check dfit basic structure

    lk = ['dprepare', 'dinput', 'dind', 'sol_x', 'jac', 'scales']
    c0 = isinstance(dfit, dict) and all([ss in dfit.keys() for ss in lk])
    if not isinstance(dfit, dict):
        msg = ("\ndfit must be a dict with at least the following keys:\n"
               + "\t- {}\n".format(lk)
               + "\t- provided: {}".format(dfit))
        raise Exception(msg)

    # ----------------
    # Identify if fit1d or fit2d

    is2d = 'nbs' in dfit['dinput'].keys()
    if is2d is True:
        if 'symmetry' not in dfit['dinput'].keys():
            msg = "dfit is a fit2d but does not have key 'symmetry'!"
            raise Exception(msg)
        if dfit['dinput']['symmetry']:
            c0 = dfit['dinput'].get('symmetry_axis', False) is False
            if c0:
                msg = "dfit is a fit2d but does not have key 'symmetry_axis'!"
                raise Exception(msg)
    else:
        phi_prof = False

    # ----------------
    # Extract dinput and dprepare (more readable)

    dinput = dfit['dinput']
    dprepare = dfit['dinput']['dprepare']

    # ----------------
    # ratio

    if ratio is None:
        ratio = False
    if ratio is not False:
        amp = ['lines', 'x']

    if ratio is not False:
        lkeys = dfit['dinput']['keys']
        if isinstance(ratio, tuple):
            ratio = [ratio]
        lc = [
            isinstance(ratio, list)
            and all([isinstance(tt, tuple) and len(tt) == 2 for tt in ratio]),
            isinstance(ratio, np.ndarray)
            and ratio.ndim == 2
            and ratio.shape[0] == 2
        ]
        msg = (
            "\nArg ratio (spectral lines magnitude ratio) must be either:\n"
            "\t- False:  no line ration computed\n"
            "\t- tuple of len=2: upper and lower keys of the lines\n"
            "\t- list of tuple of len=2: upper and lower keys pairs\n"
            "\t- np.ndarray of shape (2, N): upper keys and lower keys\n"
            f"  Available keys: {lkeys}\n"
            f"  Provided: {ratio}\n"
        )
        if not any(lc):
            warnings.warn(msg)
            ratio = False
        elif lc[0]:
            ratio = np.atleast_2d(ratio).T

        # Remove ratio using non-aviailable lines
        indokratio = np.array([
            ratio[0, ii] in lkeys and ratio[1, ii] in lkeys
            for ii in range(ratio.shape[1])
        ])
        if np.any(indokratio):
            ratio = ratio[:, indokratio]
        else:
            ratio = False

    # ----------------
    # Check / format amp, Ti, vi

    # check if double
    isdouble = dfit['dinput']['double']

    d3 = {k0: dict(v0) for k0, v0 in _D3.items()}
    lval = [
        [bck, 'bck_amp'], [bck, 'bck_rate'],
        [amp, 'amp'], [width, 'width'], [shift, 'shift'],
        [ratio, 'ratio'], [Ti, 'Ti'], [vi, 'vi'],
        [isdouble, 'dratio'], [isdouble, 'dshift'],
    ]
    for (v0, k0) in lval:
        if v0 is None or v0 is True:
            d3[k0]['requested'] = _D3[k0]['types']
        else:
            d3[k0]['requested'] = v0

    # remove non-requested
    lout = [k0 for k0, v0 in d3.items() if v0['requested'] is False]
    for k0 in lout:
        del d3[k0]

    # ----------------
    # amp, Ti, vi from d3

    lkkeys = ['amp', 'width', 'shift', 'Ti', 'vi']
    for k0 in d3.keys():

        if k0 == 'ratio':
            v0 = d3[k0]['types']
        else:
            v0 = d3[k0]['requested']

        # basic conformity check
        if isinstance(v0, str):
            v0 = [v0]
            d3[k0]['requested'] = v0

        c0 = (
            k0 != 'ratio'
            and isinstance(v0, list)
            and all([isinstance(ss, str) for ss in v0])
        )
        if not (k0 == 'ratio' or c0):
            msg = (
                f"Arg {k0} must be a list of str!\n"
                f"Provided: {v0}"
            )
            raise Exception(msg)

        # check if trying to get all/some lines and / or all/some x
        ltypes = d3[k0]['types']
        c0 = all([ss in ltypes for ss in v0]),               # all lines/x
        c1 = (
            not c0
            and 'lines' in ltypes
            and all([ss in dinput['keys'] for ss in v0]),       # some lines
        )
        c2 = (
            not c0
            and not c1
            and 'x' in ltypes
            and all([ss in dinput[k0]['keys'] for ss in v0]),   # some x
        )

        if not any([c0, c1, c2]):
            msg = (
                f"Arg {k0} elements must be either:\n"
                f"\t- 'x': return all unique {k0}\n"
                f"\t- 'lines': return {k0} for all lines (inc. duplicates)\n"
                "\t- str: a key in:\n"
                f"\t\t lines: {dinput['keys']}\n"
                f"\t\t variables: {dinput[d3[k0][1]]['keys']}\n\n"
                f"Provided: {d3[k0][0]}"
            )
            raise Exception(msg)

        if c0:
            # 'lines' and/or 'x'
            for k1 in v0:

                if k0 in lkkeys:
                    if k1 == 'lines':
                        keys = dinput['keys']
                    else:
                        keys = dinput[d3[k0]['field']]['keys']

                    d3[k0][k1] = {
                        'keys': keys,
                        'ind': np.arange(0, len(keys)),
                    }

                elif k0 != 'ratio':
                    d3[k0][k1] = {
                        'ind': np.r_[0],
                    }
                else:
                    d3[k0][k1] = {}

        else:
            if c1:
                # a selection of lines
                typ = 'lines'
                keysok = dinput['keys']
                keys = v0
                if k0 == 'amp' and ratio is not False:
                    for rr in set(ratio.ravel().tolist()):
                        if rr not in keys:
                            keys.append(rr)

            elif c2:
                # a selection of variables 'x'
                typ = 'x'
                keysok = dinput[d3[k0][1]]['keys']
                keys = v0

            d3[k0][typ] = {
                'keys': keys,
                'ind': np.array(
                    [(keysok == ss).nonzero()[0][0] for ss in keys],
                    dtype=int,
                )
            }

    # ----------------
    # phi_prof, phi_npts

    if is2d is True:
        c1 = [phi_prof is not None, phi_npts is not None]
        if all(c1):
            msg = "Arg phi_prof and phi_npts cannot be both provided!"
            raise Exception(msg)

        if phi_npts is False or phi_prof is False:
            phi_prof = False
        else:
            if not any(c1):
                phi_npts = (2*dinput['deg']-1)*(dinput['knots'].size-1) + 1
            if phi_npts is not None:
                phi_npts = int(phi_npts)
                if dfit['dinput']['symmetry'] is True:
                    phimin = (
                        np.mean(dfit['dinput']['symmetry_axis'])
                        - dprepare['domain']['phi']['minmax'][1]
                    )
                else:
                    phimin = dprepare['domain']['phi']['minmax'][0]
                phi_prof = np.linspace(
                    phimin,
                    dprepare['domain']['phi']['minmax'][1],
                    phi_npts,
                )
            else:
                phi_prof = np.atleast_1d(phi_prof).ravel()

        # vs_nbs
        if vs_nbs is None:
            vs_nbs = True
        if not isinstance(vs_nbs, bool):
            msg = "Arg vs_nbs must be a bool!"
            raise Exception(msg)

    # ----------------
    # sol_total, sol_detail, sol_pts

    if sol_pts is not None:
        if is2d is True:
            c0 = (
                isinstance(sol_pts, (list, np.ndarray))
                and len(sol_pts) == 2
                and all([isinstance(ss, np.ndarray) for ss in sol_pts])
                and sol_pts[0].shape == sol_pts[1].shape
            )
            if not c0:
                msg = (
                    "Arg sol_lamb_phi must be a tuple of 2 np.ndarray"
                    " of same shape!"
                )
                raise Exception(msg)
        else:
            c0 = isinstance(sol_pts, np.ndarray)
            if not c0:
                msg = "Arg sol_lamb must be a np.ndarray!"
                raise Exception(msg)

    if sol_total is None:
        sol_total = sol_pts is not None
    if sol_detail is None:
        sol_detail = False

    if not isinstance(sol_total, bool):
        msg = f"Arg sol_total must be a bool!\nProvided: {sol_total}"
        raise Exception(msg)
    if not isinstance(sol_detail, bool):
        msg = f"Arg sol_detail must be a bool!\nProvided: {sol_detail}"
        raise Exception(msg)

    c0 = (sol_total is True or sol_detail is True) and sol_pts is None
    if c0:
        if dprepare is None:
            sol_pts = False
        else:
            if is2d is True:
                sol_pts = [dprepare['lamb'], dprepare['phi']]
            else:
                sol_pts = dprepare['lamb']
    if any([sol_total, sol_detail]):
        assert sol_pts is not None

    return dfit, d3, sol_total, sol_detail, sol_pts, phi_prof, vs_nbs


def fit1d_extract(
    dfit1d=None,
    bck=None,
    amp=None, ratio=None,
    Ti=None, width=None,
    vi=None, shift=None,
    sol_total=None,
    sol_detail=None,
    sol_lamb=None,
):
    """
    Return a dict with extarcted data of interest

    bck_amp:    (nt,) array
    bck_rate:   (nt,) array
    amp:        (nt, namp) array
    coefs:      (nt, nlines) array
    ratio:      (nt, nratio) array
    width:      (nt, nwidth) array
    Ti:         (nt, nlines) array
    shift:      (nt, nshift) array
    vi:         (nt, nlines) array

    """

    # -------------------
    # Check format input
    (
        dfit1d, d3,
        sol_total, sol_detail, sol_lamb,
        _, _,
    ) = fit12d_get_data_checkformat(
        dfit=dfit1d,
        bck=bck,
        amp=amp, ratio=ratio,
        Ti=Ti, width=width,
        vi=vi, shift=shift,
        sol_total=sol_total,
        sol_detail=sol_detail,
        sol_pts=sol_lamb,
    )

    # Extract dprepare and dind (more readable)
    dprepare = dfit1d['dinput']['dprepare']
    dind = dfit1d['dinput']['dind']
    nspect = dprepare['data'].shape[0]

    # Prepare extract func
    def _get_values(
        k0=None,
        k1=None,
        d3=d3,
        dind=dind,
        sol_x=dfit1d['sol_x'],
        scales=dfit1d['scales'],
    ):
        ind = dind[d3[k0]['field']][k1][0, d3[k0][k1]['ind']]
        return sol_x[:, ind] * scales[:, ind]

    # -------------------
    # Prepare output

    # multiple-value, direct computation
    lk_direct = ['bck_amp', 'bck_rate', 'amp', 'width', 'shift']
    for k0 in set(lk_direct).intersection(d3.keys()):
        for k1 in set(['x', 'lines']).intersection(d3[k0].keys()):
            d3[k0][k1]['values'] = _get_values(k0=k0, k1=k1)

    # multiple-value, indirect computation
    k0 = 'Ti'
    if k0 in d3.keys():
        k1 = d3[k0]['types'][0]
        val = _get_values(k0=k0, k1=k1)
        conv = np.sqrt(scpct.mu_0*scpct.c / (2.*scpct.h*scpct.alpha))
        d3[k0][k1]['values'] = (
            val * conv * scpct.c**2
            * dfit1d['dinput']['mz'][d3[k0][k1]['ind']][None, :]
        )

    # vi
    k0 = 'vi'
    if k0 in d3.keys():
        k1 = d3[k0]['types'][0]
        val = _get_values(k0=k0, k1=k1)
        d3[k0][k1]['values'] = val * scpct.c

    # ratio
    k0 = 'ratio'
    if k0 in d3.keys():
        k1 = d3[k0]['types'][0]
        nratio = d3[k0]['requested'].shape[1]
        indup = np.r_[[
            (d3['amp']['lines']['keys'] == kk).nonzero()[0][0]
            for kk in d3['ratio']['requested'][0, :]
        ]]
        indlo = np.r_[[
            (d3['amp']['lines']['keys'] == kk).nonzero()[0][0]
            for kk in d3['ratio']['requested'][1, :]]
        ]
        val = (
            d3['amp']['lines']['values'][:, indup]
            / d3['amp']['lines']['values'][:, indlo]
        )
        lab = np.r_[
            [
                '{} / {}'.format(
                    dfit1d['dinput']['symb'][indup[ii]],
                    dfit1d['dinput']['symb'][indlo[ii]],
                )
                for ii in range(nratio)
            ]
        ]
        d3['ratio']['lines']['values'] = val
        d3['ratio']['lines']['lab'] = lab

    # double
    if dfit1d['dinput']['double'] is not False:
        double = dfit1d['dinput']['double']
        for k0 in ['dratio', 'dshift']:
            if double is True or double.get(k0) is None:
                val = _get_values(k0=k0, k1='x')
            else:
                val = np.full((nspect, 1), double[k0])
            d3[k0]['x']['values'] = val

    # -------------------
    # sol_detail and sol_tot
    sold, solt = False, False
    if any([sol_total, sol_detail]):

        (
            func_detail, func_cost,
        ) = _funccostjac.multigausfit1d_from_dlines_funccostjac(
            lamb=sol_lamb,
            indx=None,      # because dfit['sol_x' built with const]
            dinput=dfit1d['dinput'],
            dind=dind,
            jac=None,
        )[:2]

        # sol_details
        if sol_detail:
            shape = tuple(np.r_[
                nspect,
                sol_lamb.shape,
                dfit1d['dinput']['nlines'] + 1,
            ])
            sold = np.full(shape, np.nan)
            for ii in range(nspect):
                if dfit1d['validity'][ii] < 0:
                    continue
                sold[ii, ...] = func_detail(
                    dfit1d['sol_x'][ii, :],
                    scales=dfit1d['scales'][ii, :],
                    indok=None,
                    const=None,
                )

        # sol_total
        if sol_total:
            shape = tuple(np.r_[nspect, sol_lamb.shape])
            solt = np.full(shape, np.nan)
            for ii in range(nspect):
                if dfit1d['validity'][ii] < 0:
                    continue
                solt[ii, ...] = func_cost(
                    dfit1d['sol_x'][ii, :],
                    scales=dfit1d['scales'][ii, :],
                    indok=None,
                    const=None,
                    data=0.,
                )

            # Double-check consistency if possible
            if sol_detail:
                soldsum = np.nansum(sold, axis=-1)
                iok = (~np.isnan(solt)) & (~np.isnan(soldsum))
                c1 = np.allclose(solt[iok], soldsum[iok], equal_nan=True)
                if not c1:
                    msg = "Inconsistent computations detail vs total"
                    raise Exception(msg)

    dout = {
        'sol_detail': sold,
        'sol_total': solt,
        'units': 'a.u.',
        'd3': d3,
        'sol_lamb': sol_lamb,
    }
    return dout


def fit2d_extract(
    dfit2d=None,
    bck=None,
    amp=None, ratio=None,
    Ti=None, width=None,
    vi=None, shift=None,
    sol_total=None,
    sol_detail=None,
    sol_lamb_phi=None,
    phi_prof=None,
    phi_npts=None,
    lines_indphi=None,
    amp_on_bck_thresh=None,
    vs_nbs=None,
):
    """
    Return a dict with extarcted data of interest

    bck_amp:    (nt, nbs) array
    bck_rate:   (nt, nbs) array
    amp:        (nt, nbs, namp)   and/or (nt, phi_npts, namp) array
    coefs:      (nt, nbs, nlines) and/or (nt, phi_npts, nlines) array
    ratio:      (nt, nratio) array
    width:      (nt, nwidth) array
    Ti:         (nt, nlines) array
    shift:      (nt, nshift) array
    vi:         (nt, nlines) array

    """

    # -------------------
    # Check format input
    (
        dfit2d, d3,
        sol_total, sol_detail, sol_lamb_phi,
        phi_prof, vs_nbs,
    ) = fit12d_get_data_checkformat(
        dfit=dfit2d,
        bck=bck,
        amp=amp, ratio=ratio,
        Ti=Ti, width=width,
        vi=vi, shift=shift,
        sol_total=sol_total,
        sol_detail=sol_detail,
        sol_pts=sol_lamb_phi,
        phi_prof=phi_prof,
        phi_npts=phi_npts,
        vs_nbs=vs_nbs,
    )

    # lines_indphi
    if lines_indphi is None:
        lines_indphi = dfit2d['dinput']['keys']
    if isinstance(lines_indphi, str):
        lines_indphi = [lines_indphi]
    if not all([ss in dfit2d['dinput']['keys'] for ss in lines_indphi]):
        msg = (
            "Arg lines_indphi must be a list of lines keys:\n"
            f"\t- lines_indphi: {lines_indphi}\n"
            f"\t- available: {dfit2d['dinput']['keys']}\n"
        )
        raise Exception(msg)

    if amp_on_bck_thresh is None:
        amp_on_bck_thresh = 2.

    # Extract dprepare and dind (more readable)
    dprepare = dfit2d['dinput']['dprepare']
    dind = dfit2d['dinput']['dind']
    nspect = dprepare['data'].shape[0]

    # Prepare Bsplines
    nbs = dfit2d['dinput']['nbs']
    BS = BSpline(
        dfit2d['dinput']['knots_mult'],
        np.ones((nbs, 1), dtype=float),
        dfit2d['dinput']['deg'],
        extrapolate=False,
        axis=0,
    )

    # phi_prof symmetric
    if dfit2d['dinput']['symmetry'] is True:
        phi_prof2 = np.abs(
            phi_prof - np.mean(dfit2d['dinput']['symmetry_axis'])
        )
    else:
        phi_prof2 = phi_prof

    # Prepare extract func
    def _get_values(
        k0,
        k1=None,
        indtok=dfit2d['validity'] == 0,
        phi_prof2=phi_prof2,
        d3=d3,
        nspect=nspect,
        BS=BS,
        dind=dind,
        sol_x=dfit2d['sol_x'],
        scales=dfit2d['scales'],
    ):

        # For bck_amp, bck_rate and dratio, dshift
        # => need to make ind 2d !! [nbs, 1] and not 1d [nbs,]
        ind = dind[d3[k0]['field']][k1][:, d3[k0][k1]['ind']]
        assert k0 in ['dratio', 'dshift'] or ind.shape[0] == nbs

        # coefs
        shape = tuple(np.r_[nspect, ind.shape])
        coefs = np.full(shape, np.nan)
        coefs[indtok, :] = (
            sol_x[indtok, ...][:, ind] * scales[indtok, ...][:, ind]
        )

        # values at phi_prof
        shape = tuple(np.r_[nspect, phi_prof2.size, ind.shape[1]])
        val = np.full(shape, np.nan)
        for ii in indtok.nonzero()[0]:
            BS.c = coefs[ii, :, :]
            val[ii, :, :] = BS(phi_prof2)

        return coefs, val

    # multiple-value, direct computation
    lk_direct = ['bck_amp', 'bck_rate', 'amp', 'width', 'shift']
    for k0 in set(lk_direct).intersection(d3.keys()):
        for k1 in set(['x', 'lines']).intersection(d3[k0].keys()):
            d3[k0][k1]['coefs'], d3[k0][k1]['values'] = _get_values(
                k0=k0, k1=k1,
            )

    # multiple-value, indirect computation
    k0 = 'Ti'
    if k0 in d3.keys():
        k1 = d3[k0]['types'][0]
        coefs, val = _get_values(k0=k0, k1=k1)
        conv = np.sqrt(scpct.mu_0*scpct.c / (2.*scpct.h*scpct.alpha))
        mz = dfit2d['dinput']['mz'][d3[k0][k1]['ind']]
        d3[k0][k1]['coefs'] = (
            coefs * conv * scpct.c**2 * mz[None, None, :]
        )
        d3[k0][k1]['values'] = (
            val * conv * scpct.c**2 * mz[None, None, :]
        )

    # vi
    k0 = 'vi'
    if k0 in d3.keys():
        k1 = d3[k0]['types'][0]
        coefs, val = _get_values(k0=k0, k1=k1)
        d3[k0][k1]['coefs'] = coefs * scpct.c
        d3[k0][k1]['values'] = val * scpct.c

    # ratio
    k0 = 'ratio'
    if k0 in d3.keys():
        k1 = d3[k0]['types'][0]
        nratio = d3[k0]['requested'].shape[1]

        # safeguard against non-available lines
        indrequest = np.array([
            ii for ii in range(d3['ratio']['requested'].shape[1])
            if d3['ratio']['requested'][0, ii] in d3['amp']['lines']['keys']
            and d3['ratio']['requested'][1, ii] in d3['amp']['lines']['keys']
        ])

        # get indices for all ratios
        indup = np.r_[[
            (d3['amp']['lines']['keys'] == kk).nonzero()[0][0]
            for kk in d3['ratio']['requested'][0, indrequest]
        ]]
        indlo = np.r_[[
            (d3['amp']['lines']['keys'] == kk).nonzero()[0][0]
            for kk in d3['ratio']['requested'][1, indrequest]]
        ]

        # get values and labels
        val = (
            d3['amp']['lines']['values'][:, :, indup]
            / d3['amp']['lines']['values'][:, :, indlo]
        )
        lab = np.r_[
            [
                '{} / {}'.format(
                    dfit2d['dinput']['symb'][indup[ii]],
                    dfit2d['dinput']['symb'][indlo[ii]],
                )
                for ii in range(nratio)
            ]
        ]
        d3['ratio']['lines']['values'] = val
        d3['ratio']['lines']['lab'] = lab

    # double
    if dfit2d['dinput']['double'] is not False:
        double = dfit2d['dinput']['double']
        for k0 in ['dratio', 'dshift']:
            if double is True or double.get(k0) is None:
                coefs, val = _get_values(k0=k0, k1='x')
            else:
                coefs = np.full((nspect, nbs, 1), double[k0])
                val = np.full((nspect, nbs, 1), double[k0])
            d3[k0]['x']['coefs'] = coefs
            d3[k0]['x']['values'] = val

    # -------------------
    # check lines / bck ratio
    # a fit is usable if the lines amplitudes are >= background
    # using lambda0 for simplicity

    # get background for each line
    bcki = (
        d3['bck_amp']['x']['values']
        * np.exp(
            d3['bck_rate']['x']['values']
            * (
                dfit2d['dinput']['lines'][None, None, :]
                - dfit2d['dinput']['lambmin_bck']
            )
        )
    )

    amp_on_bck = d3['amp']['lines']['values'][:, :, :] / bcki

    # get indices of each line desired in lines_indphi
    iiphi = np.array([
        (dfit2d['dinput']['keys'] == ss).nonzero()[0][0]
        for ss in lines_indphi
    ])

    # check validity of amp/bck ratio
    indphi = amp_on_bck[:, :, iiphi] >= amp_on_bck_thresh
    indphi = np.all(indphi, axis=-1)

    for ii in range(nspect):
        indphi_no = np.copy(indphi[ii, ...])
        for jj in range(len(dfit2d['dinput']['valid']['ldphi'][ii])):
            indphi_no &= (
                (phi_prof < dfit2d['dinput']['valid']['ldphi'][ii][jj][0])
                | (phi_prof >= dfit2d['dinput']['valid']['ldphi'][ii][jj][1])
            )
        indphi[ii, indphi_no] = False

    if not np.any(indphi):
        msg = (
            "No usable vertical profile!\n"
            "Conditions for usability include:\n"
            f"\t- lines amplitude > = {amp_on_bck_thresh} * background\n"
            f"\t- phi in a an interval of valid data, for lines {lines_indphi}"
        )
        warnings.warn(msg)

    # ----------
    # update phi profiles
    for k0 in d3.keys():
        for k1 in {'lines', 'x'}.intersection(d3[k0].keys()):
            if 'values' in d3[k0][k1].keys():
                for jj in range(d3[k0][k1]['values'].shape[-1]):
                    d3[k0][k1]['values'][~indphi, jj] = np.nan

    # update validity according to indphi
    dfit2d['validity'][
        (dfit2d['validity'] == 0) & np.all(~indphi, axis=1)
    ] = -3

    # ----------
    # func
    (
        func_detail, func_sum,
    ) = _funccostjac.multigausfit2d_from_dlines_funccostjac(
        dinput=dfit2d['dinput'],
        dind=dind,
        jac=None,
    )[:2]

    # -------------------
    # func_tot

    def func_tot(
        lamb=None,
        phi=None,
        nspect=nspect,
        dfit2d=dfit2d,
        func_sum=func_sum,
    ):
        assert lamb.shape == phi.shape
        if dfit2d['dinput']['symmetry'] is True:
            phi = np.abs(phi - np.nanmean(dfit2d['dinput']['symmetry_axis']))

        shape = tuple(np.r_[nspect, lamb.shape])
        solt = np.full(shape, np.nan)
        for ii in range(nspect):
            if dfit2d['validity'][ii] < 0:
                continue
            # Separate and reshape output
            solt[ii, ...] = func_sum(
                dfit2d['sol_x'][ii, :],
                lamb=lamb,
                phi=phi,
                scales=dfit2d['scales'][ii, :],
                const=None,
            )
        return solt

    # -------------------
    # sol_detail and sol_tot

    sold, solt = False, False
    if sol_total or sol_detail:
        if dfit2d['dinput']['symmetry'] is True:
            sol_lamb_phi[1] = np.abs(
                sol_lamb_phi[1] - np.nanmean(dfit2d['dinput']['symmetry_axis'])
            )

        # sol_details
        if sol_detail:
            shape = tuple(np.r_[
                nspect,
                sol_lamb_phi[0].shape,
                dfit2d['dinput']['nlines'] + 1,
                nbs,
            ])
            sold = np.full(shape, np.nan)
            if dfit2d['dinput']['symmetry'] is True:
                phi = np.abs(
                    sol_lamb_phi[1]
                    - np.nanmean(dfit2d['dinput']['symmetry_axis'])
                )
            else:
                phi = sol_lamb_phi[1]

            for ii in range(nspect):
                if dfit2d['validity'][ii] < 0:
                    continue
                sold[ii, ...] = func_detail(
                    dfit2d['sol_x'][ii, :],
                    lamb=sol_lamb_phi[0],
                    phi=phi,
                    scales=dfit2d['scales'][ii, :],
                    const=None,
                    # indbs=None,
                )

        # sol_total
        if sol_total:
            solt = func_tot(
                lamb=sol_lamb_phi[0],
                phi=sol_lamb_phi[1],
            )

            # Double-check consistency if possible
            if sol_detail:
                soldsum = np.nansum(np.nansum(sold, axis=-1), axis=-1)
                iok = (~np.isnan(solt)) & (~np.isnan(soldsum))
                c1 = np.allclose(solt[iok], soldsum[iok], equal_nan=True)
                if not c1:
                    msg = "Inconsistent computations detail vs total"
                    raise Exception(msg)

    # -------------------
    # Add input args

    dout = {
        'sol_detail': sold,
        'sol_tot': solt,
        'func_tot': func_tot,
        'units': 'a.u.',
        'd3': d3,
        'amp_on_bck': amp_on_bck,
        'amp_on_bck_thresh': amp_on_bck_thresh,
        'phi_prof': phi_prof,
        'indphi': indphi,
        'lines_indphi': lines_indphi,
        'sol_lamb_phi': sol_lamb_phi,
    }

    return dout



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import datastock as ds


# tofu
from ._class07_Camera import Camera as Previous
from . import _class8_check as _check
from . import _class8_compute as _compute
from . import _class8_move as _move
from . import _class8_los_data as _los_data
from . import _class8_equivalent_apertures as _equivalent_apertures
from . import _class8_etendue_los as _etendue_los
from . import _class8_los_angles as _los_angles
from . import _class8_compute_signal as _compute_signal
from . import _class8_plot as _plot


__all__ = ['Diagnostic']


# #############################################################################
# #############################################################################
#                           Diagnostic
# #############################################################################


class Diagnostic(Previous):

    _show_in_summary = 'all'

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'diagnostic': [
            'is2d',
            'spectro',
            'camera',
        ],
    })

    def add_diagnostic(
        self,
        key=None,
        doptics=None,
        # etendue
        etendue=None,
        # config for los
        config=None,
        length=None,
        # reflections
        reflections_nb=None,
        reflections_type=None,
        key_nseg=None,
        # compute
        compute=True,
        # others
        verb=None,
        **kwdargs,
    ):

        # -----------
        # adding diag

        # check / format input
        dref, ddata, dobj = _check._diagnostics(
            coll=self,
            key=key,
            doptics=doptics,
            **kwdargs,
        )
        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

        # ---------------------
        # adding etendue / los

        key = list(dobj['diagnostic'].keys())[0]
        dopt = dobj['diagnostic'][key]['doptics']
        computable = any([len(v0['optics']) > 0 for v0 in dopt.values()])
        if compute is True and computable:
            self.compute_diagnostic_etendue_los(
                key=key,
                analytical=True,
                numerical=False,
                res=None,
                check=False,
                # los
                config=config,
                length=length,
                reflections_nb=reflections_nb,
                reflections_type=reflections_type,
                key_nseg=key_nseg,
                # bool
                verb=verb,
                plot=False,
                store='analytical',
            )

    # -----------------
    # remove
    # -----------------

    def remove_diagnostic(self, key=None, key_cam=None):
        return _check._remove(
            coll=self,
            key=key,
            key_cam=key_cam,
        )

    # -----------------
    # utilities
    # -----------------

    def get_diagnostic_ref(self, key=None, key_cam=None):
        return _check.get_ref(coll=self, key=key, key_cam=key_cam)

    def get_diagnostic_cam(self, key=None, key_cam=None):
        return _check._get_default_cam(coll=self, key=key, key_cam=key_cam)

    def get_diagnostic_data(
        self,
        key=None,
        key_cam=None,
        data=None,
        rocking_curve=None,
        **kwdargs,
        ):
        """ Return dict of data for chosen cameras

        data can be:
            'etendue'
            'amin'
            'amax'
            'tangency radius'
            'lamb'
            'lambmin'
            'lambmax'
            'res'

        """
        return _compute._get_data(
            coll=self,
            key=key,
            key_cam=key_cam,
            data=data,
            rocking_curve=rocking_curve,
            **kwdargs,
        )

    def get_diagnostic_data_concatenated(
        self,
        key=None,
        key_data=None,
        flat=None,
        ):
        """ Return concatenated data for chosen cameras


        """
        return _compute._concatenate_data(
            coll=self,
            key=key,
            key_data=key_data,
            flat=flat,
        )

    # -----------------
    # etendue computing
    # -----------------

    def compute_diagnostic_etendue_los(
        self,
        key=None,
        # parameters
        analytical=None,
        numerical=None,
        res=None,
        check=None,
        margin_par=None,
        margin_perp=None,
        # equivalent aperture
        add_points=None,
        convex=None,
        # for storing los
        config=None,
        length=None,
        reflections_nb=None,
        reflections_type=None,
        key_nseg=None,
        # bool
        verb=None,
        plot=None,
        store=None,
    ):
        """ Compute the etendue of the diagnostic (per pixel)

        Etendue (m2.sr) can be computed analytically or numerically
        If plot, plot the comparison between all computations
        If store = 'analytical' or 'numerical', overwrites the diag etendue

        """

        dcompute, store = _etendue_los.compute_etendue_los(
            coll=self,
            key=key,
            # etendue
            analytical=analytical,
            numerical=numerical,
            res=res,
            check=check,
            margin_par=margin_par,
            margin_perp=margin_perp,
            # equivalent aperture
            add_points=add_points,
            convex=convex,
            # bool
            verb=verb,
            plot=plot,
            store=store,
        )

        # compute los angles
        c0 = (
            any([np.any(np.isfinite(v0['los_x'])) for v0 in dcompute.values()])
            and store
            )
        if c0:
            _los_angles.compute_los_angles(
                coll=self,
                key=key,
                # los
                config=config,
                length=length,
                reflections_nb=reflections_nb,
                reflections_type=reflections_type,
                key_nseg=key_nseg,
                dcompute=dcompute,
            )

    # ---------------
    # utilities
    # ---------------

    def get_diagnostic_equivalent_aperture(
        self,
        key=None,
        key_cam=None,
        pixel=None,
        # inital contour
        add_points=None,
        # options
        convex=None,
        harmonize=None,
        reshape=None,
        return_for_etendue=None,
        # plot
        plot=None,
        verb=None,
        store=None,
    ):
        """"""
        return _equivalent_apertures.equivalent_apertures(
            coll=self,
            key=key,
            key_cam=key_cam,
            pixel=pixel,
            # inital contour
            add_points=add_points,
            # options
            convex=convex,
            harmonize=harmonize,
            reshape=reshape,
            return_for_etendue=return_for_etendue,
            # plot
            plot=plot,
            verb=verb,
            store=store,
        )

    # ---------------
    # wavelneght from angle
    # ---------------

    def get_diagnostic_lamb(
        self,
        key=None,
        key_cam=None,
        lamb=None,
        rocking_curve=None,
    ):
        """ Return the wavelength associated to
        - 'lamb'
        - 'lambmin'
        - 'lambmax'
        - 'res' = lamb / (lambmax - lambmin)

        """
        return _compute.get_lamb_from_angle(
            coll=self,
            key=key,
            key_cam=key_cam,
            lamb=lamb,
            rocking_curve=rocking_curve,
        )

    # ---------------
    # utilities
    # ---------------

    def get_optics_cls(self, optics=None):
        """ Return list of optics and list of their classes

        """
        return _check._get_optics_cls(coll=self, optics=optics)

    # def get_diagnostic_doptics(self, key=None):
    #     """ 
    #     Get dict of optics and corresponding classes 

    #     """
    #     return _check._get_diagnostic_doptics(coll=self, key=key)

    def get_optics_outline(
        self,
        key=None,
        add_points=None,
        mode=None,
        closed=None,
        ravel=None,
        total=None,
    ):
        """ Return the optics outline """
        return _compute.get_optics_outline(
            coll=self,
            key=key,
            add_points=add_points,
            mode=mode,
            closed=closed,
            ravel=ravel,
            total=total,
        )

    def get_optics_poly(
        self,
        key=None,
        add_points=None,
        mode=None,
        closed=None,
        ravel=None,
        total=None,
        return_outline=None,
    ):
        """ Return the optics outline """
        return _compute.get_optics_poly(
            coll=self,
            key=key,
            add_points=add_points,
            mode=mode,
            closed=closed,
            ravel=ravel,
            total=total,
            return_outline=return_outline,
        )

    def set_optics_color(self, key=None, color=None):
        return _check._set_optics_color(
            coll=self,
            key=key,
            color=color,
        )

    # -----------------
    # Moving
    # -----------------

    def move_diagnostic_to(
        self,
        key=None,
        key_cam=None,
        x=None,
        y=None,
        R=None,
        z=None,
        phi=None,
        theta=None,
        dphi=None,
        # computing
        compute=None,
        # los
        config=None,
        length=None,
        reflections_nb=None,
        reflections_type=None,
        key_nseg=None,
        # equivalent aperture
        add_points=None,
        convex=None,
        # etendue
        margin_par=None,
        margin_perp=None,
        verb=None,
    ):

        if compute is None:
            compute = True

        _move.move_to(
            self,
            key=key,
            key_cam=key_cam,
            x=x,
            y=y,
            R=R,
            z=z,
            phi=phi,
            theta=theta,
            dphi=dphi,
        )

        if compute:
            self.compute_diagnostic_etendue_los(
                key=key,
                #e etendue
                analytical=True,
                numerical=False,
                res=None,
                check=False,
                margin_par=margin_par,
                margin_perp=margin_perp,
                # equivalent aperture
                add_points=add_points,
                convex=convex,
                # los
                config=config,
                length=length,
                reflections_nb=reflections_nb,
                reflections_type=reflections_type,
                key_nseg=key_nseg,
                # bool
                verb=verb,
                plot=False,
                store='analytical',
            )

    # -----------------
    # computing
    # -----------------

    def compute_diagnostic_solid_angle(
        self,
        key=None,
        key_cam=None,
        # pts
        ptsx=None,
        ptsy=None,
        ptsz=None,
        # options
        config=None,
        visibility=None,
        # return
        return_vect=None,
        return_alpha=None,
    ):
        return _los_data.compute_solid_angles(
            coll=self,
            key=key,
            key_cam=key_cam,
            # pts
            ptsx=ptsx,
            ptsy=ptsy,
            ptsz=ptsz,
            # options
            config=config,
            visibility=visibility,
            # return
            return_vect=return_vect,
            return_alpha=return_alpha,
        )


    def compute_diagnostic_signal(
        self,
        key=None,
        key_diag=None,
        key_cam=None,
        # integrand
        key_integrand=None,
        # sampling
        method=None,
        res=None,
        mode=None,
        groupby=None,
        val_init=None,
        # signal
        brightness=None,
        # store
        store=None,
        # return
        returnas=None,
    ):
        """ Compute synthetic signal for a diagnostic and an emissivity field

        """

        return _compute_signal.compute_signal(
            coll=self,
            key=key,
            key_diag=key_diag,
            key_cam=key_cam,
            # integrand
            key_integrand=key_integrand,
            # sampling
            method=method,
            res=res,
            mode=mode,
            groupby=groupby,
            val_init=val_init,
            # signal
            brightness=brightness,
            # store
            store=store,
            # return
            returnas=returnas,
        )


    # -----------------
    # plotting
    # -----------------

    def get_diagnostic_dplot(
        self,
        key=None,
        key_cam=None,
        optics=None,
        elements=None,
        vect_length=None,
    ):
        """ Return a dict with all that's necessary for plotting

        If no optics is provided, all are returned

        elements indicate, for each optics, what should be represented:
            - 'o': outline
            - 'v': unit vectors
            - 's': summit ( = center for non-curved)
            - 'c': center (of curvature)
            - 'r': rowland circle / axis of cylinder

        returned as a dict:

        dplot = {
            'optics0': {
                'o': {
                    'x0': ...,
                    'x1': ...,
                    'x': ...,
                    'y': ...,
                    'z': ...,
                    'r': ...,
                },
                'v': {
                    'x': ...,
                    'y': ...,
                    'z': ...,
                    'r': ...,
                },
            },
        }

        """

        return _compute._dplot(
            coll=self,
            key=key,
            key_cam=key_cam,
            optics=optics,
            elements=elements,
            vect_length=vect_length,
        )

    def plot_diagnostic(
        self,
        key=None,
        key_cam=None,
        optics=None,
        elements=None,
        proj=None,
        los_res=None,
        # data plot
        data=None,
        cmap=None,
        vmin=None,
        vmax=None,
        # config
        plot_config=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        wintit=None,
        # interactivity
        color_dict=None,
        nlos=None,
        dinc=None,
        connect=None,
    ):

        return _plot._plot_diagnostic(
            coll=self,
            key=key,
            key_cam=key_cam,
            optics=optics,
            elements=elements,
            proj=proj,
            los_res=los_res,
            # data plot
            data=data,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            # config
            plot_config=plot_config,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            # interactivity
            color_dict=color_dict,
            nlos=nlos,
            dinc=dinc,
            connect=connect,
        )

    def plot_diagnostic_interpolated_along_los(
        self,
        key=None,
        key_cam=None,
        key_data_x=None,
        key_data_y=None,
        # sampling
        res=None,
        mode=None,
        segment=None,
        radius_max=None,
        # plotting
        vmin=None,
        vmax=None,
        plot=None,
        dcolor=None,
        dax=None,
    ):
        """ Compute and plot interpolated data along the los of the diagnostic

        """
        return _los_data._interpolated_along_los(
            coll=self,
            key=key,
            key_cam=key_cam,
            key_data_x=key_data_x,
            key_data_y=key_data_y,
            # sampling
            res=res,
            mode=mode,
            segment=segment,
            radius_max=radius_max,
            # plotting
            vmin=vmin,
            vmax=vmax,
            plot=plot,
            dcolor=dcolor,
            dax=dax,
            )





import itertools as itt


import numpy as np
import scipy.integrate as scpinteg
import astropy.units as asunits
import matplotlib.pyplot as plt     # DB


import datastock as ds


# ##################################################################
# ##################################################################
#               Main routine
# ##################################################################


def compute_signal(
    coll=None,
    key=None,
    key_diag=None,
    key_cam=None,
    # to be integrated
    key_integrand=None,
    # sampling
    method=None,
    res=None,
    mode=None,
    groupby=None,
    val_init=None,
    # signal
    brightness=None,
    # store
    store=None,
    # return
    returnas=None,
):

    # -------------
    # check inputs 
    # --------------

    (
        key_diag, key_cam, spectro, is2d,
        method, mode, groupby, val_init, brightness,
        key_integrand, key_mesh0,
        store, key,
        returnas,
    ) = _compute_signal_check(
        coll=coll,
        key_diag=key_diag,
        key_cam=key_cam,
        # sampling
        method=method,
        res=res,
        mode=mode,
        groupby=groupby,
        val_init=val_init,
        # signal
        brightness=brightness,
        # to be integrated
        key_integrand=key_integrand,
        # store
        store=store,
        key=key,
        # return
        returnas=returnas,
    )

    # -------------
    # prepare 
    # --------------

    shape_emiss = coll.ddata[key_integrand]['shape']

    if mode == 'abs':
        key_kR = coll.dobj['mesh'][key_mesh0]['knots'][0]
        radius_max = np.max(coll.ddata[key_kR]['data'])
    else:
        radius_max = None

    # -------------
    # compute 
    # --------------

    if method == 'los':
        dout, units = _compute_los(
            coll=coll,
            is2d=is2d,
            key_diag=key_diag,
            key_cam=key_cam,
            res=res,
            mode=mode,
            key_integrand=key_integrand,
            radius_max=radius_max,
            groupby=groupby,
            val_init=val_init,
            brightness=brightness,
        )

    else:
        pass

    # -------------
    # store 
    # --------------

    if store is True:
        _store(
            coll=coll,
            key=key,
            key_diag=key_diag,
            dout=dout,
            key_integrand=key_integrand,
            method=method,
            res=res,
            units=units,
        )

    # -------------
    # return 
    # --------------

    if returnas is dict:
        return dout


# ##################################################################
# ##################################################################
#               STORE
# ##################################################################


def _store(
    coll=None,
    key=None,
    key_diag=None,
    dout=None,
    units=None,
    # synthetic signal
    key_integrand=None,
    method=None,
    res=None,
    # retrofit
    key_matrix=None,
):

    # ---------
    # check

    lc = [key_integrand is not None, key_matrix is not None]
    if np.sum(lc) != 1:
        msg = "Please provide key_integrand xor key_matrix"
        raise Exception(msg)

    typ = 'retrofit' if key_integrand is None else 'synthetic'

    # ---------
    # prepare

    doptics = coll._dobj['diagnostic'][key_diag]['doptics']
    dsig = coll._dobj['diagnostic'][key_diag].get('dsignal')
    if dsig is None:
        dsig = {}

    lkc = list(dout.keys())
    lksig = [f'{key}_{k0}' for k0 in lkc]

    # ----------
    # build dict

    dsig.update({
        key: {
            'type': typ,
            'camera': lkc,
            'data': lksig,
            # synthetic
            'integrand': key_integrand,
            'method': method,
            'res': res,
            # retrofit
            'geom matrix': key_matrix,
        },
    })

    # ----------
    # add data

    for ii, k0 in enumerate(lkc):

        # add data
        coll.add_data(
            key=lksig[ii],
            data=dout[k0]['data'],
            ref=dout[k0]['ref'],
            units=units,
        )

    coll._dobj['diagnostic'][key_diag]['dsignal'] = dsig


# ##################################################################
#               CHECK
# ##################################################################


def _compute_signal_check(
    coll=None,
    key=None,
    key_diag=None,
    key_cam=None,
    # sampling
    method=None,
    res=None,
    mode=None,
    groupby=None,
    val_init=None,
    # signal
    brightness=None,
    # to be integrated
    key_integrand=None,
    # store
    store=None,
    # return
    returnas=None,
):

    # key_diag, key_cam
    key_diag, key_cam = coll.get_diagnostic_cam(key=key_diag, key_cam=key_cam)
    spectro = coll.dobj['diagnostic'][key_diag]['spectro']
    is2d = coll.dobj['diagnostic'][key_diag]['is2d']

    # method
    method = ds._generic_check._check_var(
        method, 'method',
        types=str,
        default='los',
        allowed=['los', 'vos'],
    )

    # mode
    mode = ds._generic_check._check_var(
        mode, 'mode',
        types=str,
        default='abs',
        allowed=['abs', 'rel'],
    )

    # groupby
    groupby = ds._generic_check._check_var(
        groupby, 'groupby',
        types=int,
        default=200,
    )

    # brightness
    brightness = ds._generic_check._check_var(
        brightness, 'brightness',
        types=bool,
        default=False,
    )

    # key_integrand
    lok = [
        k0 for k0, v0 in coll.ddata.items()
        if v0.get('bsplines') is not None
    ]
    key_integrand = ds._generic_check._check_var(
        key_integrand, 'key_integrand',
        types=str,
        allowed=lok,
    )

    # key_mesh0
    key_bs = coll.ddata[key_integrand]['bsplines']
    key_mesh = coll.dobj['bsplines'][key_bs]['mesh']
    mtype = coll.dobj['mesh'][key_mesh]['type']
    if mtype == 'polar':
        key_mesh0 = coll.dobj['mesh'][key_mesh]['submesh']
    else:
        key_mesh0 = key_mesh

    # val_init
    val_init = ds._generic_check._check_var(
        val_init, 'val_init',
        default=np.nan,
        allowed=[np.nan, 0.]
    )

    # store
    store = ds._generic_check._check_var(
        store, 'store',
        types=bool,
        default=True,
    )

    # key
    lsig = list(coll.dobj['diagnostic'][key_diag].get('dsignal', {}).keys())
    lout = list(coll.ddata.keys()) + lsig
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        default=f'{key_diag}_synth',
        excluded=lout,
    )

    # returnas
    returnas = ds._generic_check._check_var(
        returnas, 'returnas',
        default=False if store is True else dict,
        allowed=[dict, False],
    )

    return (
        key_diag, key_cam, spectro, is2d,
        method, mode, groupby, val_init, brightness,
        key_integrand, key_mesh0,
        store, key,
        returnas,
    )


# ##################################################################
# ##################################################################
#               LOS
# ##################################################################


def _compute_los(
    coll=None,
    is2d=None,
    key_diag=None,
    key_cam=None,
    res=None,
    mode=None,
    key_integrand=None,
    radius_max=None,
    groupby=None,
    val_init=None,
    brightness=None,
):

    # ----------------
    # loop on cameras

    dout = {}
    doptics = coll.dobj['diagnostic'][key_diag]['doptics']
    for k0 in key_cam:

        npix = coll.dobj['camera'][k0]['dgeom']['pix_nb']
        key_los = doptics[k0]['los']

        ngroup = npix // groupby
        if groupby * ngroup < npix:
            ngroup += 1

        # -----------------------
        # loop on group of pixels (to limit memory footprint)

        for ii in range(ngroup):

            i0 = ii*groupby
            i1 = min((ii + 1)*groupby, npix)
            ni = i1 - i0

            R, Z, length = coll.sample_rays(
                key=key_los,
                res=res,
                mode=mode,
                segment=None,
                ind_flat=np.arange(i0, i1),
                radius_max=radius_max,
                concatenate=True,
                return_coords=['R', 'z', 'ltot'],
            )

            inan = np.isnan(R)
            inannb = np.r_[-1, inan.nonzero()[0]]
            nnan = inan.sum()
            assert nnan == ni, f"{nnan} vs {ni}"
            iok = ~inan

            # -------------
            # interpolate

            datai, units, refi = coll.interpolate_profile2d(
                key=key_integrand,
                R=R,
                Z=Z,
                grid=False,
                radius_vs_time=None,
                azone=None,
                t=None,
                indt=None,
                indt_strict=None,
                indbs=None,
                details=False,
                reshape=None,
                crop=None,
                nan0=True,
                val_out=np.nan,
                return_params=False,
                store=False,
            )

            axis = refi.index(None)
            if ii == 0:
                shape = list(datai.shape)
                shape[axis] = npix
                data = np.full(shape, val_init)
                ref = list(refi)

            # ------------
            # integrate

            iok2 = np.isfinite(datai)
            sli0 = [slice(None) for aa in range(len(refi))]
            for jj in range(nnan):

                # slice datai
                indi = np.arange(inannb[jj]+1, inannb[jj+1])
                sli0[axis] = indi
                slii = tuple(sli0)
                if not np.any(iok2[slii]):
                    continue

                # set nan to 0 for integration
                dataii = datai[slii]
                dataii[~iok2[slii]] = 0.

                # slice data
                ind = i0 + jj
                sli0[axis] = ind
                sli = tuple(sli0)

                # if jj in [50, 51]:
                    # plt.figure();
                    # plt.subplot(1,2,1)
                    # plt.plot(dataii)
                    # plt.subplot(1,2,2)
                    # plt.plot(dataii.T)
                    # plt.gcf().suptitle(f"jj = {jj}", size=12)

                # integrate
                data[sli] = scpinteg.simpson(
                    dataii,
                    x=length[indi],
                    axis=axis,
                )

        # --------------
        # post-treatment

        # brightness
        if brightness is False:
            ketend = doptics[k0]['etendue']
            etend = coll.ddata[ketend]['data']
            sh_etend = [-1 if aa == axis else 1 for aa in range(len(refi))]
            data *= etend.reshape(sh_etend)

        # reshape if 2d
        if is2d:
            sh_data = list(data.shape)
            sh_data[axis] = coll.dobj['camera'][k0]['dgeom']['shape']
            sh_data = tuple(np.r_[
                sh_data[:axis], sh_data[axis], sh_data[axis+1:]
            ].astype(int))
            data = data.reshape(sh_data)

        # set ref
        ref[axis] = coll.dobj['camera'][k0]['dgeom']['ref']
        ref = tuple(np.r_[ref[:axis], ref[axis], ref[axis+1:]])

        # fill dout
        dout[k0] = {
            'data': data,
            'ref': ref,
        }

    # -----
    # units

    units0 = coll.ddata[key_integrand]['units']
    units = units0 * asunits.m
    if brightness is False:
        units = units * coll.ddata[ketend]['units']

    return dout, units


# ##################################################################
# ##################################################################
#               VOS
# ##################################################################


def _compute_vos(
    coll=None,
    key_cam=None,
    res=None,
    mode=None,
    key_integrand=None,
):

    dout = None

    return dout



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: cdivision=True
#

ctypedef struct SortedSet:
    double val
    SortedSet* next
    SortedSet* last

cdef void insert(SortedSet** head, double val) nogil
cdef int count(SortedSet* head, double val) nogil
cdef void free_ss(SortedSet** head) nogil



#!/usr/bin/env python

import os
import argparse

import numpy as np

_save = True
_here = os.path.abspath(os.path.dirname(__file__))
_Exp, _Cls, _name = os.path.split(__file__)[1].split('_')[:3]
assert not any([any([ss in s for ss in ['Notes','.']])
               for s in [_Exp, _Cls, _name]])

def get_notes():

    # Notes from creoView (-X,Z,Y)
    notes = {'C': np.r_[2.465, 0.],
             'r_in': 3.162/2., # r_out for later use (thick)
             'r_out': 3.292/2.}
    return notes

def make_Poly(save=_save, path=_here):

    notes = get_notes()

    C = notes['C']
    nP = 100
    theta = np.linspace(0.,2*np.pi, nP, endpoint=False)
    P = np.array([C[0]+notes['r_out']*np.cos(theta),
                  C[1]+notes['r_out']*np.sin(theta)])

    if save:
        cstr = '%s_%s_%s'%(_Exp,_Cls,_name)
        pathfilext = os.path.join(path, cstr+'_V0.txt')
        np.savetxt(pathfilext, P)
    return P, notes



if __name__=='__main__':

    # Parse input arguments
    msg = 'Launch creation of polygons txt from bash'
    parser = argparse.ArgumentParser(description = msg)

    parser.add_argument('-save', type=bool, help='save ?', default=_save)
    parser.add_argument('-path', type=str, help='saving path ?', default=_here)

    args = parser.parse_args()

    # Call wrapper function
    make_Poly(save=args.save, path=args.path)




# Built-in
import os
import shutil
import requests
import warnings
import csv

# Common
import numpy as np


__all__ = [
    'step01_search_online_by_wavelengthA',
    'clear_cache',
]


# Check whether a local .tofu/ repo exists
_URL = 'https://physics.nist.gov/PhysRefData/ASD/lines_form.html'
_URL_SEARCH_PRE = 'https://physics.nist.gov/cgi-bin/ASD/'
_URL_SOURCE = 'https://physics.nist.gov/cgi-bin/ASBib1/get_ASBib_ref.cgi?'
_CUSTOM = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
_CUSTOM = os.path.join(_CUSTOM, 'scripts', 'tofucustom.py')
_CREATE_CUSTOM = True
_CACHE_FROM = True
_CACHE_TO = True
_CACHE_UPDATE = True
_CACHE_INFO = False


_DOP = {
    'ion': {'str': 'lines1.pl?spectra=', 'cache': 'A'},
    'lambmin': {'str': 'low_w=', 'cache': 'B'},
    'lambmax': {'str': 'upp_w=', 'cache': 'C'},
    'wav_observed': {'str': 'show_obs_wl=1', 'def': True, 'cache': 'D'},
    'wav_calculated': {'str': 'show_calc_wl=1', 'def': True, 'cache': 'E'},
    'transitions_allowed': {'str': 'allowed_out=1', 'def': True, 'cache': 'F'},
    'transitions_forbidden': {
        'str': 'forbid_out=1', 'def': True, 'cache': 'G',
    },
    'info_ref': {'str': 'bibrefs=1', 'def': True, 'cache': 'H'},
    'info_conf': {'str': 'conf_out=on', 'def': True, 'cache': 'H'},
    'info_term': {'str': 'term_out=on', 'def': True, 'cache': 'J'},
    'info_energy': {'str': 'enrg_out=on', 'def': True, 'cache': 'K'},
    'info_J': {'str': 'J_out=on', 'def': True, 'cache': 'L'},
    'info_g': {'str': 'g_out=on', 'def': True, 'cache': 'M'},
}


_LTYPES = [int, float, np.int_, np.float_]


_DCERTIFICATES_BUNDLE = {
    'ITER': {
        'host': 'iter.org',
        'bund': '/etc/pki/ca-trust/extracted/openssl/ca-bundle.trust.crt',
    },
}


# #############################################################################
# #############################################################################
#                           Utility functions
# #############################################################################


def _getcharray(
    ar,
    col=None, sep='  ', line='-', just='l',
    verb=True, returnas=str,
):
    """ Format and return char array (for pretty printing) """
    c0 = ar is None or len(ar) == 0
    if c0:
        return ''
    ar = np.array(ar, dtype='U')

    if ar.ndim == 1:
        ar = ar.reshape((1, ar.size))

    # Get just len
    nn = np.char.str_len(ar).max(axis=0)
    if col is not None:
        if len(col) not in ar.shape:
            msg = ("len(col) should be in np.array(ar, dtype='U').shape:\n"
                   + "\t- len(col) = {}\n".format(len(col))
                   + "\t- ar.shape = {}".format(ar.shape))
            raise Exception(msg)
        if len(col) != ar.shape[1]:
            ar = ar.T
            nn = np.char.str_len(ar).max(axis=0)
        nn = np.fmax(nn, [len(cc) for cc in col])

    # Apply to array
    fjust = np.char.ljust if just == 'l' else np.char.rjust
    out = np.array([sep.join(v) for v in fjust(ar, nn)])

    # Apply to col
    if col is not None:
        arcol = np.array([col, [line*n for n in nn]], dtype='U')
        arcol = np.array([sep.join(v) for v in fjust(arcol, nn)])
        out = np.append(arcol, out)

    if verb is True:
        print('\n'.join(out))
    if returnas is str:
        return '\n'.join(out)
    elif returnas is np.ndarray:
        return ar


def _get_PATH_LOCAL(strict=True):
    pfe = os.path.join(os.path.expanduser('~'), '.tofu', 'nist2tofu')
    if os.path.isdir(pfe):
        return pfe
    else:
        if strict is True:
            return None
        else:
            return pfe


def _get_cache_url(dop=None):

    # direct
    cache_url = '_'.join([
        '{}{}'.format(
            v0['cache'],
            v0['str'].split('=')[1]
        )
        for k0, v0 in dop.items()
        if 'submit' not in v0['str'] and v0['val'] is not False
    ])
    return cache_url


def _get_totalurl(
    element=None,
    charge=None,
    ion=None,
    lambmin=None,
    lambmax=None,
    wav_observed=None,
    wav_calculated=None,
    transitions_allowed=None,
    transitions_forbidden=None,
    info_ref=None,
    info_conf=None,
    info_term=None,
    info_energy=None,
    info_J=None,
    info_g=None,
    dop=dict(_DOP),
):

    # ----------
    # check inputs

    # ion, element, charge
    lc = [
        ion is not None,
        charge is not None or element is not None,
    ]
    if all(lc) and ion != '{}{}+'.format(element, charge):
        msg = (
            "\nArg ion, element and charge cannot be provided together!\n"
            + "Please povide either ion xor (element and charge)!\n"
            + "  You provided:\n"
            + "\t- element: {}\n".format(element)
            + "\t- charge: {}\n".format(charge)
            + "\t- ion: {}\n".format(ion)
        )
        raise Exception(msg)
    if lc[1]:
        c0 = (
            isinstance(element, str)
            and (
                charge is None
                or isinstance(charge, int)
            )
        )
        if not c0:
            msg = (
                "\nArg element must be a str and charge must be a int!\n"
                + "  You provided:\n"
                + "\t- element: {}\n".format(type(element))
                + "\t- charge: {}\n".format(type(charge))
            )
            raise Exception(msg)
        ion = element
        if charge is not None:
            ion = "{}{}+".format(element, charge)

    # ion
    if ion is None:
        ion = ''
    c0 = (
        isinstance(ion, str)
        or (
            (isinstance(ion, list) or isinstance(ion, np.ndarray))
            and all([isinstance(ss, str) for ss in ion])
        )
    )
    if not c0:
        msg = (
            "\nArg ion must be a str or list of such!\n"
            + "  e.g.: 'Ar16' or ['ar', 'w44+']\n"
            + "  You provided:\n"
            + "\t- ion: {}".format(ion)
        )
        raise Exception(msg)
    if isinstance(ion, str):
        ion = ion.replace('+', '')
    else:
        ion = '%3B'.join([ss.replace('+', '') for ss in ion])

    # lamb
    dlamb = {'lambmin': lambmin, 'lambmax': lambmax}
    for k0, v0 in dlamb.items():
        if v0 is None:
            dlamb[k0] = ''
        else:
            c0 = type(v0) in _LTYPES
            if not c0:
                msg = (
                    "Arg {} must be a float!\n".format(k0)
                    + "\t- provided: {}".format(type(v0))
                )
                raise Exception(msg)
            dlamb[k0] = '{}'.format(v0)

    # options
    dop['ion']['val'] = ion
    dop['ion']['str'] = 'lines1.pl?spectra={}'.format(ion)
    dop['lambmin']['val'] = dlamb['lambmin']
    dop['lambmin']['str'] = 'low_w={}'.format(dlamb['lambmin'])
    dop['lambmax']['val'] = dlamb['lambmax']
    dop['lambmax']['str'] = 'upp_w={}'.format(dlamb['lambmax'])
    dop['wav_observed']['val'] = wav_observed
    dop['wav_calculated']['val'] = wav_calculated
    dop['transitions_allowed']['val'] = transitions_allowed
    dop['transitions_forbidden']['val'] = transitions_forbidden
    dop['info_ref']['val'] = info_ref
    dop['info_conf']['val'] = info_conf
    dop['info_term']['val'] = info_term
    dop['info_energy']['val'] = info_energy
    dop['info_J']['val'] = info_J
    dop['info_g']['val'] = info_g

    for k0, v0 in dop.items():
        if k0 in ['ion', 'lambmin', 'lambmax']:
            continue
        if v0['val'] is None:
            dop[k0]['val'] = dop[k0]['def']
        if not isinstance(dop[k0]['val'], bool):
            msg = (
                "Arg {} must be a bool!\n".format(k0)
                + "\t- provided: {}".format(type(dop[k0]['val']))
            )
            raise Exception(msg)

    # ---------
    # build url

    lsearchurl = [
        'lines1.pl?spectra={}'.format(ion),
        'limits_type=0',
        'low_w={}'.format(dlamb['lambmin']),
        'upp_w={}'.format(dlamb['lambmax']),
        'unit=0',       # 0 = A, 1 = nm, 2 = um
        'submit=Retrieve+Data',
        'de=0',
        'format=2',         # 0 = html, 1 = ASCII, 2 = CSV, 3 = tab-delimited
        'line_out=0',
        'en_unit=1',        # 0 = cm^-1, 1 = eV, 2 = Rydberg
        'output=0',         # 0 = all results, 1 = display by page
        'bibrefs=1',        # show bibliographic ref. (remove otherwise)
        'page_size=15',     # nb of results per page (if output=1)
        'show_obs_wl=1',    # show observed wavelengths (remove otherwise)
        'show_calc_wl=1',   # show calculated wavelength (remove otherwise)
        'unc_out=1',        # show uncertainty
        'order_out=0',      # 0 = sort by wavelength, 1 = by Multiplet
        'max_low_enrg=',
        'show_av=3',        # 0-2 = vac/air (2000, 10000,  both), 3 = vacuum
        'max_upp_enrg=',
        'tsb_value=0',
        'min_str=',
        'A_out=0',
        'intens_out=on',
        'max_str=',
        'allowed_out=1',    # show allowed transitions (E1), remove otherwise
        'forbid_out=1',     # show forbidden transitions (M1, E2...), remove...
        'min_accur=',
        'min_intens=',
        'conf_out=on',      # level info: show configuration (remove...)
        'term_out=on',      # level info: show term (remove...)
        'enrg_out=on',      # level info: show energy (remove...)
        'J_out=on',         # level info: show J (remove...)
        'g_out=on',         # level info: show g (remove...)
    ]

    for k0, v0 in dop.items():
        if v0['val'] is False:
            lsearchurl.remove(v0['str'])

    # total url (online source)
    total_url = '{}{}'.format(_URL_SEARCH_PRE, '&'.join(lsearchurl))

    # cache url (loacal cahed source)
    path_local = _get_PATH_LOCAL(strict=False)
    cache_url = os.path.join(
        path_local, 'ASD', _get_cache_url(dop=dop) + '.csv',
    )
    path_local = _get_PATH_LOCAL(strict=True)
    return cache_url, total_url, path_local, dop


# #############################################################################
# #############################################################################
#                           SSL certificates handling
# #############################################################################


def _try_request_handle_ITER_SSL_step01(url=None):
    try:
        resp = requests.get(url)
    except requests.exceptions.SSLError as err:
        if 'certificate' in str(err):
            lbund = [
                vv['bund'] for kk, vv in _DCERTIFICATES_BUNDLE.items()
                if os.uname()[1].endswith(vv['host'])
            ]
            if len(lbund) == 1:
                resp = requests.get(url, verify=lbund[0])
            else:
                msg = (
                    str(err)
                    + "\n\nLooks like a certificate error occured!\n"
                    + "=> try changing the certificate bundle of requests\n"
                    + "=> ask your admin which certificate bundle to use!"
                )
                raise Exception(msg)
        else:
            raise err
    except Exception as err:
        raise err
    return resp


def _try_request_handle_ITER_SSL_step02(url=None, pfe=None):
    try:
        with requests.get(url, stream=True) as rr:
            rr.raise_for_status()
            with open(pfe, 'wb') as ff:
                for chunk in rr.iter_content(chunk_size=8192):
                    # filter-out keep-alive new chunks
                    if chunk:
                        ff.write(chunk)
                        # ff.flush()
    except requests.exceptions.SSLError as err:
        if 'certificate' in str(err):
            lbund = [
                vv['bund'] for kk, vv in _DCERTIFICATES_BUNDLE.items()
                if os.uname()[1].endswith(vv['host'])
            ]
            if len(lbund) == 1:
                with requests.get(url, stream=True, verify=lbund[0]) as rr:
                    rr.raise_for_status()
                    with open(pfe, 'wb') as ff:
                        for chunk in rr.iter_content(chunk_size=8192):
                            # filter-out keep-alive new chunks
                            if chunk:
                                ff.write(chunk)
                                # ff.flush()
            else:
                msg = (
                    str(err)
                    + "\n\nLooks like a certificate error occured!\n"
                    + "=> try changing the certificate bundle of requests\n"
                    + "=> ask your admin which certificate bundle to use!"
                )
                raise Exception(msg)
        else:
            raise err
    except Exception as err:
        raise err


# #############################################################################
# #############################################################################
#                           csv parsing
# #############################################################################


def _csv_parser(
    pfe=None, url=None, path_local=None,
    create_custom=None,
    dlines0=None,
):

    # Download url
    if url is not None:
        if path_local is None:
            if create_custom is True:
                os.system('python ' + _CUSTOM)
                path_local = _get_PATH_LOCAL()
            else:
                path = os.path.join(
                    os.path.expanduser('~'), '.tofu', 'nist2tofu',
                )
                msg = (
                    "You do not seem to have a local ./tofu repository\n"
                    + "tofu uses that local repository to store user-specific "
                    + "data and downloads\n"
                    + "In particular, openadas files are downloaded in:\n"
                    + "\t{}\n".format(path)
                    + "  => to set-up your local .tofu repo, run (terminal):\n"
                    + "\ttofu custom"
                )
                raise Exception(msg)

        try:
            _try_request_handle_ITER_SSL_step02(url=url, pfe=pfe)
        except Exception as err:
            msg = (
                str(err)
                + "\n\nFile could not be downloaded:\n"
                + "\t{}\n".format(url)
                + "  => Maybe check internet connection?"
            )
            raise Exception(msg)

    # nb to exclude if dlines0
    if dlines0 is not None:
        c0 = (
            isinstance(dlines0, dict)
            and all([
                isinstance(ss, str)
                and isinstance(vv, dict)
                and 'lambda0' in vv.keys()
                for ss, vv in dlines0.items()
            ])
        )
        if not c0:
            msg = "Arg dlines0 is not a dlines dict!"
            raise Exception(msg)
        lnb = [
            int(kk[len('nist_'):].split('_')[0]) for kk in dlines0.keys()
            if kk.startswith('nist_')
        ]
        if len(lnb) > 0:
            ii = np.nanmax(lnb) + 1
        else:
            ii = 0
        ltrans = (['conf_i', 'term_i', 'J_i'], ['conf_k', 'term_k', 'J_k'])
    else:
        ii = 0

    # Read file
    dout = {}
    lkout = ['Unnamed: 0', 'Aki(s^-1)', 'Acc', 'Unnamed: 22', '']
    ok = True
    with open(pfe, mode='r') as csv_file:
        csv_reader = csv.DictReader(csv_file, delimiter=',')
        line_count = 0
        for row in csv_reader:
            if line_count == 0:
                line_count += 1
            else:
                # remove useless char
                rowi = {
                    k0: v0.replace('=', '').replace('"', '')
                    for k0, v0 in row.items()
                    if k0 not in lkout
                }

                if ok is True:
                    key = 'nist_{:03}'.format(ii)
                    dout[key] = rowi
                    line_count += 1
                    ii += 1

    return dout


# #############################################################################
# #############################################################################
#                          online search
# #############################################################################


def step01_search_online_by_wavelengthA(
    element=None,
    charge=None,
    ion=None,
    lambmin=None,
    lambmax=None,
    wav_observed=None,
    wav_calculated=None,
    transitions_allowed=None,
    transitions_forbidden=None,
    info_ref=None,
    info_conf=None,
    info_term=None,
    info_J=None,
    info_g=None,
    cache_from=None,
    cache_info=None,
    return_dout=None,
    return_dsources=None,
    return_url=None,
    verb=None,
    create_custom=None,
    format_for_DataStock=None,
    dsource0=None,
    dlines0=None,
):
    """ Perform an online freeform search on https://open.adas.ac.uk

    Pass searchstr to the online freeform search
    Prints the results (if verb=True)

    Optionally print and return the result as a pandas dataframe

    example
    -------
        >>> import tofu as tf
        >>> tf.nist2tofu.step01_search_online_by_wavelengthA(
            lambmin=3.94,
            lambmax=4,
            ion='ar',
        )
    """

    # ----------
    # Check input
    if cache_from is None:
        cache_from = _CACHE_FROM
    if cache_info is None:
        cache_info = _CACHE_INFO
    if return_dout is None:
        return_dout = False
    if return_dsources is None:
        return_dsources = False
    if return_url is None:
        return_url = False
    if verb is None:
        verb = True
    if create_custom is None:
        create_custom = _CREATE_CUSTOM
    if format_for_DataStock is None:
        format_for_DataStock = False

    # ----------
    # get search url
    cache_url, total_url, path_local, dop = _get_totalurl(
        element=element,
        charge=charge,
        ion=ion,
        lambmin=lambmin,
        lambmax=lambmax,
        wav_observed=wav_observed,
        wav_calculated=wav_calculated,
        transitions_allowed=transitions_allowed,
        transitions_forbidden=transitions_forbidden,
        info_ref=info_ref,
        info_conf=info_conf,
        info_term=info_term,
        info_J=info_J,
        info_g=info_g,
    )

    # ----------
    # load from cache or online
    c0 = (
        cache_from is True
        and path_local is not None
        and os.path.isfile(cache_url)
    )
    loaded_from_cache = False
    if c0:
        dout = _csv_parser(url=None, pfe=cache_url, dlines0=dlines0)
        if cache_info is True:
            msg = "Loaded from cache:\n\t{}".format(cache_url)
            print(msg)
        loaded_from_cache = True
    else:
        dout = _csv_parser(
            url=total_url, pfe=cache_url,
            path_local=path_local,
            create_custom=create_custom,
            dlines0=dlines0,
        )

    # Trivial case
    lcol = list(list(dout.values())[0].keys())
    if '<!DOCTYPE html' in lcol:
        if format_for_DataStock is True:
            return None, None
        else:
            lv = [
                (None, return_dout),
                (None, return_dsources),
                (None, return_url),
            ]
            out = tuple([vv[0] for vv in lv if vv[1] is True])
            return out

    # ----------
    # Complete source
    if 'element' not in lcol or 'sp_num' not in lcol:
        if dop['ion']['val'] == 'H':
            for k0 in dout.keys():
                dout[k0]['element'] = 'H'
                dout[k0]['sp_num'] = '1'
        else:
            msg = (
                "Unknown case"
            )
            raise Exception(msg)

    dsources = None
    if 'line_ref' in lcol:
        lsources = sorted(set([
            (
                v0['line_ref'],
                v0['element'],
                v0['sp_num'],
            )
            for k0, v0 in dout.items()
            if v0['line_ref'] != ''
        ]))
        dsources = _get_dsources(lsources)

    # ----------
    # print
    if verb is True:
        col = list(list(dout.values())[0].keys())
        char = np.array(
            [[k0] + [v0[kk] for kk in col] for k0, v0 in dout.items()]
        )
        col = ['key'] + col
        arr = _getcharray(char, col=col,
                          sep='  ', line='-', just='l',
                          returnas=str, verb=verb)

    # ----------
    # return
    if format_for_DataStock is True:
        return _format_for_DataStock(
            dout=dout, dsources=dsources, dlines0=dlines0,
        )
    else:
        url = cache_url if loaded_from_cache else total_url
        lv = [
            (dout, return_dout),
            (dsources, return_dsources),
            (url, return_url),
        ]
        out = tuple([vv[0] for vv in lv if vv[1] is True])
        if len(out) == 0:
            out = None
        elif len(out) == 1:
            out = out[0]
        return out


# #############################################################################
# #############################################################################
#                          sources
# #############################################################################


def _get_dsources(lsources):
    dsources = {}
    for ss in lsources:
        dbid = ss[0].split('c')
        if len(dbid) > 1:
            dbid, code = dbid[0][1:], 'c'+dbid[1]
        else:
            dbid, code = dbid[0][1:], ''
        largs = [
            'db=el',
            'db_id={}'.format(dbid),
            'comment_code={}'.format(code),
            'element={}'.format(ss[1]),
            'spectr_charge={}'.format(ss[2]),
            'type=',
        ]
        url = _URL_SOURCE + '&'.join(largs)

        try:
            resp = _try_request_handle_ITER_SSL_step01(url=url).text

            # title
            if resp.count('<title>') == resp.count('</title>') == 1:
                i0 = resp.index('<title>')+len('<title>')
                i1 = resp.index('</title>')
                title = resp[i0:i1]
            else:
                msg = "title could not be identified"
                raise Exception(msg)

            # First author
            char = (
                '<a id="aa" title="Click to search for all papers of this'
                + ' author" target="_blank" href='
            )
            if resp.count(char) > 0:
                i0 = resp.index(char) + len(char)
                i1 = resp[i0:].index('</a>')
                author = resp[i0:i0+i1].replace('"', '')
                author = author.split(';')[-1]
            else:
                msg = "author could not be identified"
                raise Exception(msg)

            # journal
            char = (
                '<a id="aj" title="Click to open the journal'
            )
            extralen = len('s online archive" href=') + 1
            if resp.count(char) >= 1:
                i0 = resp.index(char) + len(char) + extralen
                i1 = resp[i0:].index('</a>')
                jour = resp[i0:i0+i1].replace('"', '')
                jour = jour.split('>')[-1]
                refs = resp[i0+i1+len('</a>'):]
                refs = refs[:refs.index(')')]
                irefs, year = refs.split('(')
                vol = refs[refs.index('<b>')+len('<b>'):refs.index('</b>')]
            else:
                msg = "jour, year, vol could not be identified"
                raise Exception(msg)

            # url
            char = (
                '<a id="at" title="Click to open the article in the '
                + 'journal online archive" target="_blank" href='
            )
            if resp.count(char) == 1:
                i0 = resp.index(char) + len(char)
                i1 = resp[i0:].index('>')
                url = resp[i0:i0+i1].replace('"', '')
            else:
                msg = "url could not be identified"
                raise Exception(msg)

            longi = "{} et al., {}, {}, {}, {}".format(
                author, title, jour, vol, year,
            )
            dsources[ss[0]] = {'long': longi, 'url': url}
        except Exception as err:
            dsources[ss[0]] = {'long': ss}

    return dsources


# #############################################################################
# #############################################################################
#                          format for DataStock
# #############################################################################


def _extract_one_line(
    dout=None, dsources=None,
    k0=None, key=None, lamb0=None,
    dlines=None, lcol=None,
    dlines0=None,
):
    ion = '{}{}+'.format(dout[k0]['element'].title(), dout[k0]['sp_num'])
    ls = [('conf_i', 'term_i', 'J_i'), ('conf_k', 'term_k', 'J_k')]
    if all([ss in lcol for ss in ls[0]]) and all([ss in lcol for ss in ls[1]]):
        trans = (
            '{} {} {}'.format(*[dout[k0][kk] for kk in ls[0]]),
            '{} {} {}'.format(*[dout[k0][kk] for kk in ls[1]]),
        )
    if 'line_ref' in lcol and dout[k0]['line_ref'] in dsources.keys():
        source = dout[k0]['line_ref']
    else:
        source = 'unknown'
    nn, tt = key.split('_')[1:]
    symbol = 'n{}{}'.format(nn, tt[0])

    # check for prexisting line
    ok = True
    if dlines0 is not None:
        lk = [
            kk for kk, vv in dlines0.items()
            if kk.startswith('nist_')
            and kk.endswith(key.split('_')[-1])
            if vv['transition'] == trans
            and vv['ion'] == ion
        ]
        if len(lk) == 0:
            ok = True
        else:
            ok = False

    if ok is True:
        dlines[key] = {
            'ion': ion,
            'lambda0': lamb0,
            'source': source,
            'transition': trans,
            'symbol': symbol,
        }
        if source == 'unknown' and 'unknown' not in dsources.keys():
            dsources['unknown'] = {'long': 'unknown', 'url': 'None'}
    return dlines, dsources


def _format_for_DataStock(
    dout=None,
    dsources=None,
    dsource0=None,
    dlines0=None,
):
    dlines = {}
    lcol = list(list(dout.values())[0].keys())
    for k0 in dout.keys():
        kl = 'obs_wl_vac(A)'
        c0 = kl in lcol and dout[k0][kl] != ''
        if c0:
            key = '{}_obs'.format(k0)
            lamb0 = float(dout[k0][kl])*1e-10
            dlines, dsources = _extract_one_line(
                dout=dout, dsources=dsources,
                k0=k0, key=key, lamb0=lamb0,
                dlines=dlines, lcol=lcol,
                dlines0=dlines0,
            )
        kl = 'ritz_wl_vac(A)'
        c0 = kl in lcol and dout[k0][kl] != ''
        if c0:
            key = '{}_calc'.format(k0)
            lamb0 = float(dout[k0][kl])*1e-10
            dlines, dsources = _extract_one_line(
                dout=dout, dsources=dsources,
                k0=k0, key=key, lamb0=lamb0,
                dlines=dlines, lcol=lcol,
                dlines0=dlines0,
            )

    if dsource0 is not None:
        dsources = {
            kk: vv for kk, vv in dsources.items()
            if kk not in dsource0.keys()
        }
    return dlines, dsources


# #############################################################################
# #############################################################################
#                          cache
# #############################################################################


def clear_cache():
    """ Delete all nist files downloaded in your ~/.tofu/ directory """
    path_local = _get_PATH_LOCAL()
    if path_local is None:
        return
    pathASD = os.path.join(path_local, 'ASD')
    lf = [
        os.path.join(pathASD, ff)
        for ff in os.listdir(pathASD)
        if os.path.isfile(os.path.join(pathASD, ff))
        and ff.endswith('.csv')
    ]
    for ff in lf:
        os.remove(ff)



# -*- coding: utf-8 -*-


import warnings


import numpy as np


import datastock as ds


__all__ = ['compute_los_angles']


# ##################################################################
# ##################################################################
#                       Main
# ##################################################################


def compute_los_angles(
    coll=None,
    key=None,
    # los
    dcompute=None,
    # for storing los
    config=None,
    length=None,
    reflections_nb=None,
    reflections_type=None,
    key_nseg=None,
    **kwdargs,
):

    # ------------
    # check inputs

    # key
    lok = list(coll.dobj.get('diagnostic', {}))
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )
    is2d = coll.dobj['diagnostic'][key]['is2d']

    for key_cam, v0 in dcompute.items():
        
        if v0['los_x'] is None or not np.any(np.isfinite(v0['los_x'])):
            continue
        
        # klos
        klos = f'{key}_{key_cam}_los'
    
        # ref
        ref = coll.dobj['camera'][key_cam]['dgeom']['ref']
    
        # ------------
        # add los
    
        cx2, cy2, cz2 = coll.get_camera_cents_xyz(key=key_cam)
    
        coll.add_rays(
            key=klos,
            start_x=cx2,
            start_y=cy2,
            start_z=cz2,
            vect_x=v0['los_x'],
            vect_y=v0['los_y'],
            vect_z=v0['los_z'],
            ref=ref,
            diag=key,
            key_cam=key_cam,
            config=config,
            length=length,
            reflections_nb=reflections_nb,
            reflections_type=reflections_type,
            key_nseg=key_nseg,
        )
    
        coll._dobj['diagnostic'][key]['doptics'][key_cam]['los'] = klos

        # ------------
        # for spectro => estimate angle variations

        if v0.get('dlos_x') is not None:

            angmin = np.full(v0['cx'].size, np.nan)
            angmax = np.full(v0['cx'].size, np.nan)
    
            ptsvect = coll.get_optics_reflect_ptsvect(key=v0['kref'])
    
            for ii in range(v0['cx'].size):
                
                if not v0['iok'][ii]:
                    continue
                
                angles = ptsvect(
                    pts_x=v0['cx'][ii],
                    pts_y=v0['cy'][ii],
                    pts_z=v0['cz'][ii],
                    vect_x=v0['dlos_x'][ii, ...],
                    vect_y=v0['dlos_y'][ii, ...],
                    vect_z=v0['dlos_z'][ii, ...],
                    strict=True,
                    return_x01=False,
                )[6]
    
                angmin[ii] = np.nanmin(angles)
                angmax[ii] = np.nanmax(angles)
    
            if is2d:
                angmin = angmin.reshape(v0['shape0'])
                angmax = angmax.reshape(v0['shape0'])
    
            # ddata
            kamin = f'{key}_{key_cam}_amin'
            kamax = f'{key}_{key_cam}_amax'
            ddata = {
                kamin: {
                    'data': angmin,
                    'ref': ref,
                    'dim': 'angle',
                    'quant': 'angle',
                    'name': 'alpha',
                    'units': 'rad',
                },
                kamax: {
                    'data': angmax,
                    'ref': ref,
                    'dim': 'angle',
                    'quant': 'angle',
                    'name': 'alpha',
                    'units': 'rad',
                },
            }
            coll.update(ddata=ddata)
    
            coll._dobj['diagnostic'][key]['doptics'][key_cam]['amin'] = kamin
            coll._dobj['diagnostic'][key]['doptics'][key_cam]['amax'] = kamax


# -*- coding: utf-8 -*-


import warnings
import itertools as itt


import numpy as np
import matplotlib.colors as mcolors
import datastock as ds


# #############################################################################
# #############################################################################
#                           Diagnostics
# #############################################################################


def _diagnostics_check(
    coll=None,
    key=None,
    doptics=None,
    stack=None,
):

    # ----
    # key

    key = ds._generic_check._obj_key(
        d0=coll.dobj.get('diagnostic', {}), short='diag', key=key,
    )

    # ------
    # doptics

    # preliminary checks
    if isinstance(doptics, str):
        doptics = {doptics: []}
    if isinstance(doptics, (list, tuple)):
        doptics = {doptics[0]: list(doptics[1:])}

    err = False
    if not isinstance(doptics, dict):
        err = True

    c0 = any([
        not isinstance(k0, str) or not isinstance(v0, (str, list))
        for k0, v0 in doptics.items()
        ])
    if c0:
        err = True

    # detailed checks
    dkout = None
    if err is False:

        for k0, v0 in doptics.items():
            if isinstance(v0, str):
                doptics[k0] = [v0]

        lcam = list(coll.dobj.get('camera', {}).keys())
        lap = list(coll.dobj.get('aperture', {}).keys())
        lfilt = list(coll.dobj.get('filter', {}).keys())
        lcryst = list(coll.dobj.get('crystal', {}).keys())
        lgrat = list(coll.dobj.get('grating', {}).keys())

        lop = lap + lfilt + lcryst + lgrat

        dkout = {
            k0: [k1 for k1 in v0 if k1 not in lop]
            for k0, v0 in doptics.items()
            if k0 not in lcam
            or any([k1 not in lop for k1 in v0])
            }
        if len(dkout) > 0:
            err = True

    if err:
        msg = (
            f"diag '{key}': arg doptics must be a dict with:\n"
            "\t- keys: key to existing camera\n"
            "\t- values: list of existing optics (aperture, filter, crystal)\n"
            )
        if dkout is not None and len(dkout) > 0:
            lstr = [f"\t- {k0}: {v0}" for k0, v0 in dkout.items()]
            msg += "Wrong key / value pairs:\n" + "\n".join(lstr)
        else:
            msg += f"\nProvided:\n{doptics}"
        raise Exception(msg)

    # -----------------
    # types of camera

    lcam = list(doptics.keys())
    types = [coll.dobj['camera'][k0]['dgeom']['type'] for k0 in lcam]

    if len(set(types)) > 1:
        msg = (
            f"diag '{key}': all cameras must be of the same type (1d or 2d)!\n"
            f"\t- cameras: {lcam}\n"
            f"\t- types: {types}"
            )
        raise Exception(msg)

    is2d = types[0] == '2d'

    # -------------------------------------------------
    # check all optics are on good side of each camera

    for cam in lcam:

        dgeom_cam = coll.dobj['camera'][cam]['dgeom']
        last_ref = cam
        last_ref_cls = 'camera'
        for oo in doptics[cam]:

            if oo in lap:
                cls = 'aperture'
            elif oo in lfilt:
                cls = 'filter'
            elif oo in lcryst:
                cls = 'crystal'
            else:
                cls = 'grating'

            dgeom = coll.dobj[cls][oo]['dgeom']

            px, py, pz = coll.get_optics_poly(key=oo)

            dgeom_lastref = coll.dobj[last_ref_cls][last_ref]['dgeom']

            if (last_ref == cam and is2d) or last_ref != cam:
                cent = dgeom_lastref['cent']
                nin = dgeom_lastref['nin']

                iout = (
                    (px - cent[0])*nin[0]
                    + (py - cent[1])*nin[1]
                    + (pz - cent[2])*nin[2]
                ) <= 0
                if np.any(iout):
                    msg = (
                        f"diag '{key}':\n"
                        f"The following points of aperture '{oo}' are on the wrong"
                        f"side of lastref '{cam}':\n"
                        f"{iout.nonzero()[0]}"
                    )
                    raise Exception(msg)

            else:
                assert last_ref == cam and not is2d

                cx, cy, cz = dgeom_cam['cents']
                cx = coll.ddata[cx]['data'][None, :]
                cy = coll.ddata[cy]['data'][None, :]
                cz = coll.ddata[cz]['data'][None, :]

                if dgeom_cam['parallel']:
                    ninx, niny, ninz = dgeom_cam['nin']
                else:
                    ninx, niny, ninz = dgeom_cam['nin']
                    ninx = coll.ddata[ninx]['data'][None, :]
                    niny = coll.ddata[niny]['data'][None, :]
                    ninz = coll.ddata[ninz]['data'][None, :]

                iout = (
                    (px[:, None] - cx)*ninx
                    + (py[:, None] - cy)*niny
                    + (pz[:, None] - cz)*ninz
                ) <= 0
                if np.any(iout):
                    msg = (
                        f"The following points of {cls} '{oo}' are on the wrong"
                        f"side of camera '{cam}':\n"
                        f"{np.unique(iout.nonzero()[0])}"
                    )
                    warnings.warn(msg)

            # update last_ref ?
            if cls in ['crystal', 'grating']:
                last_ref = oo
                last_ref_cls = cls

    # -----------------
    # is spectro

    dspectro = {
        k0 : any([
            k1 in coll.dobj.get('crystal', {}).keys()
            or k1 in coll.dobj.get('grating', {}).keys()
            for k1 in v0
            ])
        for k0, v0 in doptics.items()
    }

    lc = [
        all([v0 for v0 in dspectro.values()]),
        all([not v0 for v0 in dspectro.values()]),
    ]
    if not any(lc):
        msg = (
            f"diag '{key}' must be either all spectro or all non-spectro!\n"
            + "\n".join([f"\t- {k0}: {v0}" for k0, v0 in dspectro.items()])
            )
        raise Exception(msg)

    spectro = lc[0] is True

    # -----------------
    # rearrange doptics

    doptics2 = {}
    for k0, v0 in doptics.items():
        doptics2[k0] = {
            'camera': k0,
            'los': None,
            'vos': None,
            'etendue': None,
            'etend_type': None,
            'amin': None,
            'amax': None,
            }

        doptics2[k0]['optics'], doptics2[k0]['cls'] = _get_optics_cls(
            coll=coll,
            optics=v0,
        )

    # -----------
    # ispectro

    if spectro:
        for k0, v0 in doptics2.items():
            doptics2[k0]['ispectro'] = [
                ii for ii, cc in enumerate(v0['cls'])
                if cc in ['grating', 'crystal']
            ]

    # -----------------
    # stack

    stack = ds._generic_check._check_var(
        stack, 'stack',
        types=str,
        default='horizontal',
        allowed=['horizontal', 'vertical'],
        )

    return key, lcam, doptics2, is2d, spectro,stack


def _diagnostics(
    coll=None,
    key=None,
    doptics=None,
    stack=None,
    **kwdargs,
):

    # ------------
    # check inputs

    key, lcam, doptics, is2d, spectro, stack = _diagnostics_check(
        coll=coll,
        key=key,
        doptics=doptics,
        stack=stack,
    )

    # --------
    # dobj

    dobj = {
        'diagnostic': {
            key: {
                'camera': lcam,
                'doptics': doptics,
                'ncam': len(doptics),
                # 'npix tot.': np.sum(),
                'is2d': is2d,
                'spectro': spectro,
                'stack': stack,
            },
        },
    }

    # -----------
    # kwdargs

    if len(kwdargs) > 0:
        for k0, v0 in kwdargs.items():
            if not isinstance(k0, str):
                continue
            elif k0 in dobj['diagnostic'][key].keys():
                continue
            else:
                dobj['diagnostic'][key][k0] = v0

    return None, None, dobj


# ##################################################################
# ##################################################################
#                           get ref
# ##################################################################


def _get_default_cam(coll=None, key=None, key_cam=None):

    # ----------
    # key

    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )
    is2d = coll.dobj['diagnostic'][key]['is2d']
    # spectro = coll.dobj['diagnostic'][key]['spectro']

    # -----------------------
    # key_cam (only 1 if is2d)

    lok = list(coll.dobj['diagnostic'][key]['doptics'].keys())
    if is2d:
        # 2d: can only select one camera at a time
        key_cam_def = [coll.dobj['diagnostic'][key]['camera'][0]]
    else:
        key_cam_def = None

    if isinstance(key_cam, str):
        key_cam = [key_cam]

    key_cam = ds._generic_check._check_var_iter(
        key_cam, 'key_cam',
        types=list,
        types_iter=str,
        allowed=lok,
        default=key_cam_def,
    )

    return key, key_cam


def get_ref(coll=None, key=None):

    # ---------
    # check key

    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )

    # -----------
    # return

    cam = coll.dobj['diagnostic'][key]['optics'][0]
    return coll.dobj['camera'][cam]['dgeom']['ref']





# ##################################################################
# ##################################################################
#                       get optics
# ##################################################################


def _get_optics_cls(coll=None, optics=None):

    # ---------
    # check key

    lcls = ['camera', 'aperture', 'filter', 'crystal', 'grating']

    if isinstance(optics, str):
        optics = [optics]

    lok = list(itt.chain.from_iterable([
        list(coll.dobj.get(cc, {}).keys())
        for cc in lcls
    ]))
    optics = ds._generic_check._check_var_iter(
        optics, 'optics',
        types=list,
        types_iter=str,
        allowed=lok,
    )

    # -----------
    # return

    optics_cls = []
    derr = {}
    for ii, oo in enumerate(optics):
        lc = [cc for cc in lcls if oo in coll.dobj.get(cc, {}).keys()]
        if len(lc) == 1:
            optics_cls.append(lc[0])
        else:
            derr[oo] = lc

    # --------
    # error

    if len(derr) > 0:
        msg = (
            f"The following have no / several classes:\n"
            + "\n".join([f"\t- {k0}: {v0}" for k0, v0 in derr.items()])
        )
        raise Exception(msg)

    return optics, optics_cls


# def _get_diagnostic_doptics(coll=None, key=None):

#     # ---------
#     # check key

#     lok = list(coll.dobj.get('diagnostic', {}).keys())
#     key = ds._generic_check._check_var(
#         key, 'key',
#         types=str,
#         allowed=lok,
#     )

#     # ------------------
#     # optics and classes

#     doptics = {}
#     for k0, v0 in coll.dobj['diagnostic'][key]['doptics']:
#         doptics[k0]['camera'] = k0
#         doptics[k0]['optics'], doptics[k0]['cls'] = _get_optics_cls(
#             coll=coll,
#             optics=v0,
#         )

#     # -----------
#     # ispectro

#     if coll.dobj['diagnostic'][key]['spectro']:
#         for k0, v0 in doptics.items():
#             doptics[k0]['ispectro'] = [
#                 ii for ii, cc in enumerate(v0['cls'])
#                 if cc in ['grating', 'crystal']
#             ]

#     return doptics


# ##################################################################
# ##################################################################
#                           set color
# ##################################################################


def _set_optics_color(
    coll=None,
    key=None,
    color=None,
):

    # ------------
    # check inputs

    # key
    lk = ['aperture', 'filter', 'crystal', 'grating', 'camera', 'diagnostic']
    dk = {
        k0: list(coll.dobj.get(k0, {}).keys())
        for k0 in lk
    }
    lok = itt.chain.from_iterable([vv for vv in dk.values()])

    if isinstance(key, str):
        key = [key]

    key = ds._generic_check._check_var_iter(
        key, 'key',
        types=(list, tuple),
        types_iter=str,
        allowed=lok,
    )

    # color
    if color is None:
        color = 'k'

    if not mcolors.is_color_like(color):
        msg = (
            f"Arg color for '{key}' must be a matplotlib color!\n"
            f"Provided: {color}\n"
        )
        raise Exception(msg)

    color = mcolors.to_rgba(color)

    # ------------
    # set color

    for k0 in key:
        cls = [k1 for k1, v1 in dk.items() if k0 in v1][0]
        if cls == 'diagnostic':
            for k2 in coll._dobj[cls][k0]['optics']:
                cls2 = [k1 for k1, v1 in dk.items() if k2 in v1][0]
                coll._dobj[cls2][k2]['dmisc']['color'] = color
        else:
            coll._dobj[cls][k0]['dmisc']['color'] = color

    return


# ##################################################################
# ##################################################################
#                           remove
# ##################################################################

def _remove(
    coll=None,
    key=None,
    key_cam=None,
):

    # ------------
    # check inputs

    lok = list(coll.dobj.get('diagnostic', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )
    doptics = coll.dobj['diagnostic'][key]['doptics']

    if key_cam is None:
        key_cam = coll.dobj['diagnostic'][key]['camera']

    else:
        if isinstance(key_cam, str):
            key_cam = [key_cam]

        key_cam = ds._generic_check._check_var_iter(
            key_cam, 'key_cam',
            types=list,
            types_iter=str,
            allowed=coll.dobj['diagnostic'][key]['camera'],
        )

    # ------------
    # list data

    lkd = ['etendue']
    ld = []
    for k0 in lkd:
        for k1 in key_cam:
            if doptics[k1][k0] is not None:
                ld.append(doptics[k1][k0])

    # -----------
    # remove data

    if len(ld) > 0:
        coll.remove_data(ld, propagate=True)

    # ----------
    # remove los

    for k1 in key_cam:
        if doptics[k1]['los'] is not None:
            coll.remove_rays(key=doptics[k1]['los'])

    # -----------
    # remove diag

    if len(key_cam) == len(doptics):
        del coll._dobj['diagnostic'][key]



# -*- coding: utf-8 -*-


import itertools as itt


import numpy as np
import scipy.interpolate as scpinterp
import datastock as ds


from . import _class2_check as _check


# ###############################################################
# ###############################################################
#                   sample
# ###############################################################


def _sample(
    coll=None,
    key=None,
    key_cam=None,
    res=None,
    mode=None,
    segment=None,
    ind_flat=None,
    radius_max=None,
    concatenate=None,
    return_coords=None,
):

    # ------------
    # check inputs

    (
        res, mode, concatenate,
        segment, ind_flat, radius_max,
        return_coords, out_xyz, out_k,
    ) = _sample_check(
        res=res,
        mode=mode,
        concatenate=concatenate,
        segment=segment,
        ind_flat=ind_flat,
        radius_max=radius_max,
        return_coords=return_coords,
    )

    # -----------
    # prepare
    # -----------

    # ------------------------
    # get points of interest

    pts_x, pts_y, pts_z = coll.get_rays_pts(key=key, key_cam=key_cam)
    npts = pts_x.shape[0]
    npix = np.prod(pts_x.shape[1:])
    i0 = np.arange(0, npts)

    # segment
    if segment is not None:
        segment[segment < 0] = npts - 1 + segment[segment < 0]
        iseg = np.r_[segment, segment[-1] + 1]

    # ind_flat
    if ind_flat is not None:
        ind_flat[ind_flat < 0] = npix - 1 + ind_flat[ind_flat < 0]

        pts_x = np.reshape(pts_x, (npts, -1))[:, ind_flat]
        pts_y = np.reshape(pts_y, (npts, -1))[:, ind_flat]
        pts_z = np.reshape(pts_z, (npts, -1))[:, ind_flat]

    # length
    length_orig = np.sqrt(
        np.diff(pts_x, axis=0)**2
        + np.diff(pts_y, axis=0)**2
        + np.diff(pts_z, axis=0)**2
    )
    zeros = np.zeros(tuple(np.r_[1, length_orig.shape[1:]]), dtype=float)
    length0 = np.cumsum(np.concatenate((zeros, length_orig), axis=0), axis=0)
    length1 = np.concatenate((length_orig, length_orig[-1:, ...]), axis=0)

    # changes to pts
    if radius_max is not None:
        pts_x, pts_y, pts_z, i0 = coll.get_rays_intersect_radius(
            key=key,
            key_cam=key_cam,
            segment=segment,
            axis_radius=radius_max,
            lim_to_segments=True,
            return_pts=True,
            return_itot=True,
        )[3:]

        # ind_flat
        if ind_flat is not None:
            pts_x = np.reshape(pts_x, (pts_x.shape[0], -1))[:, ind_flat]
            pts_y = np.reshape(pts_y, (pts_x.shape[0], -1))[:, ind_flat]
            pts_z = np.reshape(pts_z, (pts_x.shape[0], -1))[:, ind_flat]
            i0 = np.reshape(i0, (pts_x.shape[0], -1))[:, ind_flat]

        length_rad = np.sqrt(
            np.diff(pts_x, axis=0)**2
            + np.diff(pts_y, axis=0)**2
            + np.diff(pts_z, axis=0)**2
        )

        if segment is not None:
            length0 = length0[iseg, ...]
            length1 = length1[iseg, ...]

    else:
        if segment is not None:
            i0 = i0[iseg]
            pts_x = pts_x[i0, ...]
            pts_y = pts_y[i0, ...]
            pts_z = pts_z[i0, ...]
            length0 = length0[iseg, ...]
            length1 = length1[iseg, ...]

        length_rad = length0[1:, ...]

    # -----------
    # compute
    # -----------

    # -------------------------
    # prepare sampling indices

    # rel
    if mode == 'rel':

        # make sure npts allow to describe all integer indices
        nptsi = (i0[-1] - i0[0]) * int(np.ceil(1./res))
        N = int(np.ceil((nptsi - 1) / (i0[-1] - i0[0])))
        nptsi = N * (i0[-1] - i0[0]) + 1
        itot = np.linspace(i0[0], i0[-1], nptsi)

        # interpolate
        if out_xyz:
            pts_x = scpinterp.interp1d(
                i0,
                pts_x,
                kind='linear',
                axis=0,
            )(itot)
            pts_y = scpinterp.interp1d(
                i0,
                pts_y,
                kind='linear',
                axis=0,
            )(itot)
            pts_z = scpinterp.interp1d(
                i0,
                pts_z,
                kind='linear',
                axis=0,
            )(itot)

        if 'l' in return_coords or 'ltot' in return_coords:
            i1 = np.floor(itot).astype(int)
            i1[i1 == length1.shape[0] - 1] -= 1
            length = length1[i1, ...]
            lengthtot = length0[i1, ...]

    # abs => for pts.ndim >= 3 (2d cameras and above), flattened list
    else:

        iok = np.isfinite(pts_x)
        nn = np.ceil(length_rad / res).astype(int)

        lpx, lpy, lpz, itot, llen, lentot = [], [], [], [], [], []
        for ind in itt.product(*[range(ss) for ss in pts_x.shape[1:]]):

            sli = tuple([slice(None)] + list(ind))
            if not np.any(iok[sli]):
                continue

            if radius_max is None:
                i0i = i0[iok[sli]]
            else:
                i0i = i0[sli]

            itoti = np.concatenate(tuple(
                [
                    np.linspace(
                        i0i[jj], i0i[jj+1], nn[tuple(np.r_[jj, ind])] + 1,
                        )[:-1]
                    for jj in range(i0i.size - 1)
                ]
                + [[i0i[-1]]]
                ))
            itot.append(itoti)

            # interpolate
            if out_xyz:
                lpx.append(scpinterp.interp1d(
                    i0i,
                    pts_x[sli],
                    kind='linear',
                    axis=0,
                )(itoti))

                lpy.append(scpinterp.interp1d(
                    i0i,
                    pts_y[sli],
                    kind='linear',
                    axis=0,
                )(itoti))

                lpz.append(scpinterp.interp1d(
                    i0i,
                    pts_z[sli],
                    kind='linear',
                    axis=0,
                )(itoti))

            if 'l' in return_coords or 'ltot' in return_coords:
                i1 = np.floor(itoti).astype(int)
                i1[i1 == length0.shape[0] - 1] -= 1
                llen.append(length1[sli][i1])
                lentot.append(length0[sli][i1])

        if out_xyz:
            pts_x, pts_y, pts_z = lpx, lpy, lpz
        if 'l' in return_coords or 'ltot' in return_coords:
            length = llen
            lengthtot = lentot

    # -------------------------------------
    # optional concatenation (for plotting)

    if concatenate is True:
        if mode == 'rel':
            shape = tuple(np.r_[np.r_[1], pts_x.shape[1:]])
            nan = np.full(shape, np.nan)
            if out_k:
                itot2 = np.full(pts_x.shape, np.nan)
                for ii in range(pts_x.shape[0]):
                    itot2[ii, ...] = itot[ii]
                itot = np.concatenate((itot2, nan), axis=0).T.ravel()
            if out_xyz:
                pts_x = np.concatenate((pts_x, nan), axis=0).T.ravel()
                pts_y = np.concatenate((pts_y, nan), axis=0).T.ravel()
                pts_z = np.concatenate((pts_z, nan), axis=0).T.ravel()
            if 'l' in return_coords or 'ltot' in return_coords:
                length = np.concatenate((length, nan), axis=0).T.ravel()
                lengthtot = np.concatenate((lengthtot, nan), axis=0).T.ravel()


        else:
            if out_k:
                itot = np.concatenate(
                    tuple([np.append(pp, np.nan) for pp in itot])
                    )
            if out_xyz:
                pts_x = np.concatenate(
                    tuple([np.append(pp, np.nan) for pp in pts_x])
                    )
                pts_y = np.concatenate(
                    tuple([np.append(pp, np.nan) for pp in pts_y])
                    )
                pts_z = np.concatenate(
                    tuple([np.append(pp, np.nan) for pp in pts_z])
                    )
            if 'l' in return_coords or 'ltot' in return_coords:
                length = np.concatenate(
                    tuple([np.append(pp, np.nan) for pp in length])
                    )
                lengthtot = np.concatenate(
                    tuple([np.append(pp, np.nan) for pp in lengthtot])
                    )

    # -------------
    # return

    if out_k:
        if concatenate is True or mode == 'rel':
            kk = itot - np.floor(itot)
            kk[itot == np.nanmax(itot)] = 1.
        else:
            kk = [ii - np.floor(ii) for ii in itot]

    lout = []
    for cc in return_coords:
        if cc == 'x':
            lout.append(pts_x)

        elif cc == 'y':
            lout.append(pts_y)

        elif cc == 'z':
            lout.append(pts_z)

        elif cc == 'R':
            if concatenate is True or mode == 'rel':
                lout.append(np.hypot(pts_x, pts_y))
            else:
                lout.append([
                    np.hypot(px, py)
                    for px, py in zip(pts_x, pts_y)
                ])

        elif cc == 'phi':
            if concatenate is True or mode == 'rel':
                lout.append(np.arctan2(pts_y, pts_x))
            else:
                lout.append([
                    np.arctan2(py, px)
                    for px, py in zip(pts_x, pts_y)
                ])

        elif cc == 'ang_vs_ephi':
            if concatenate is True or mode == 'rel':
                phi = np.arctan2(pts_y, pts_x)
                ux = np.diff(pts_x, axis=0)
                uy = np.diff(pts_y, axis=0)
                ux = np.concatenate((ux[0:1, ...], ux), axis=0)
                uy = np.concatenate((uy[0:1, ...], uy), axis=0)
                vn = np.sqrt(ux**2 + uy**2)
                ux = ux / vn
                uy = uy / vn
                lout.append(np.arccos(-np.sin(phi)*ux + np.cos(phi)*uy))
            else:
                lout.append([])
                for px, py in zip(pts_x, pts_y):
                    phi = np.arctan2(py, px)
                    ux = np.diff(px, axis=0)
                    uy = np.diff(py, axis=0)
                    ux = np.concatenate((ux[0:1, ...], ux), axis=0)
                    uy = np.concatenate((uy[0:1, ...], uy), axis=0)
                    vn = np.sqrt(ux**2 + uy**2)
                    ux = ux / vn
                    uy = uy / vn
                    lout[-1].append(np.arccos(-np.sin(phi)*ux + np.cos(phi)*uy))

        elif cc == 'itot':
            lout.append(itot)

        elif cc == 'k':
            lout.append(kk)

        elif cc == 'l':
            if concatenate is True or mode == 'rel':
                lout.append(kk*length)
            else:
                lout.append([
                    kki * ll
                    for kki, ll in zip(kk, length)
                ])

        elif cc == 'ltot':
            if concatenate is True or mode == 'rel':
                # import matplotlib.pyplot as plt
                # plt.figure()
                # plt.subplot(1,4,1)
                # plt.plot(kk)
                # plt.subplot(1,4,2)
                # plt.plot(length)
                # plt.subplot(1,4,3)
                # plt.plot(lengthtot)
                # plt.subplot(1,4,4)
                # plt.plot(kk*length + lengthtot)
                # import pdb; pdb.set_trace()     # DB

                lout.append(kk*length + lengthtot)
            else:
                lout.append([
                    kki * ll1 + ll0
                    for kki, ll1, ll0 in zip(kk, length, lengthtot)
                ])

    return lout


def _sample_check(
    res=None,
    mode=None,
    segment=None,
    ind_flat=None,
    radius_max=None,
    concatenate=None,
    return_coords=None,
):

    # res
    res = ds._generic_check._check_var(
        res, 'res',
        types=float,
        default=0.1,
        sign='> 0',
    )

    # mode
    mode = ds._generic_check._check_var(
        mode, 'mode',
        types=str,
        default='rel',
        allowed=['rel', 'abs'],
    )

    # concatenate
    concatenate = ds._generic_check._check_var(
        concatenate, 'concatenate',
        types=bool,
        default=False,
    )

    # segment
    if segment is not None:
        segment = np.atleast_1d(segment).astype(int).ravel()

    # ind_flat
    if ind_flat is not None:
        ind_flat = np.atleast_1d(ind_flat).astype(int).ravel()

    # tangency_radius_max
    if radius_max is not None:
        radius_max = ds._generic_check._check_var(
            radius_max, 'radius_max',
            types=(float, int),
            sign='> 0',
        )

        if mode == 'rel':
            msg = "radius_max can only be used with mode='abs'!"
            raise Exception(msg)

    # return_coords
    if isinstance(return_coords, str):
        return_coords = [return_coords]

    lok_k = ['k', 'l', 'ltot', 'itot']
    lok_xyz = ['x', 'y', 'z', 'R', 'phi', 'ang_vs_ephi']
    return_coords = ds._generic_check._check_var_iter(
        return_coords, 'return_coords',
        types=list,
        types_iter=str,
        default=['x', 'y', 'z'],
        allowed=lok_k + lok_xyz,
    )

    out_xyz = any([ss in return_coords for ss in lok_xyz])
    out_k = any([ss in return_coords for ss in lok_k])

    return (
        res, mode, concatenate,
        segment, ind_flat, radius_max,
        return_coords, out_xyz, out_k,
    )


# ###############################################################
# ###############################################################
#                  tangency radius
# ###############################################################


def _tangency_radius_check(
    coll=None,
    key=None,
    key_cam=None,
    quantity=None,
    segment=None,
    lim_to_segments=None,
    axis_pt=None,
    axis_vect=None,
):

    # key
    key = _check._check_key(coll=coll, key=key, key_cam=key_cam)

    # quantity
    quantity = ds._generic_check._check_var(
        quantity, 'quantity',
        types=str,
        default='tangency radius',
        allowed=['alpha', 'tangency radius', 'length'],
    )

    # lim_to_segments
    lim_to_segments = ds._generic_check._check_var(
        lim_to_segments, 'lim_to_segments',
        types=bool,
        default=False,
    )

    # segment
    if segment is not None:
        segment = np.atleast_1d(segment).astype(int).ravel()

    # --------
    # tangency radius-specific

    if quantity == 'tangency radius':
        # axis_pt
        if axis_pt is None:
            axis_pt = [0., 0., 0.]

        axis_pt = ds._generic_check._check_flat1darray(
            axis_pt, 'axis_pt',
            dtype=float,
            size=3,
            can_be_None=False,
        )

        # axis_vect
        if axis_vect is None:
            axis_vect = [0., 0., 1.]

        axis_vect = ds._generic_check._check_flat1darray(
            axis_vect, 'axis_vect',
            dtype=float,
            size=3,
            norm=True,
            can_be_None=False,
        )

    return key, quantity, segment, lim_to_segments, axis_pt, axis_vect


def _tangency_radius_prepare(
    coll=None,
    key=None,
    quantity=None,
    segment=None,
    axis_pt=None,
    axis_vect=None,
):

    pts_x, pts_y, pts_z = coll.get_rays_pts(key=key)
    i0 = np.arange(0, pts_x.shape[0])

    # select segment
    if segment is not None:
        npts = pts_x.shape[0]
        segment[segment < 0] = npts - 1 + segment[segment < 0]

        iseg = np.r_[segment, segment[-1] + 1]
        i0 = i0[iseg]
        pts_x = pts_x[iseg, :]
        pts_y = pts_y[iseg, :]
        pts_z = pts_z[iseg, :]


    # define vectors
    ABx = np.diff(pts_x, axis=0)
    ABy = np.diff(pts_y, axis=0)
    ABz = np.diff(pts_z, axis=0)

    if quantity == 'tangency radius':
        AOx = axis_pt[0] - pts_x[:-1, ...]
        AOy = axis_pt[1] - pts_y[:-1, ...]
        AOz = axis_pt[2] - pts_z[:-1, ...]

        # --------
        # get kk

        # Eq:
        # k^2 |AB x v|^2 + 2k((AB.v)(AO.v) - AB.AO) + |AO x v|^2 = dist^2
        # 
        # minimum:
        # k |AB x v|^2 + (AB.v)(AO.v) - AB.AO = 0

        ABvn2 = (
            (ABy*axis_vect[2] - ABz*axis_vect[1])**2
            + (ABz*axis_vect[0] - ABx*axis_vect[2])**2
            + (ABx*axis_vect[1] - ABy*axis_vect[0])**2
            )
        AOvn2 = (
            (AOy*axis_vect[2] - AOz*axis_vect[1])**2
            + (AOz*axis_vect[0] - AOx*axis_vect[2])**2
            + (AOx*axis_vect[1] - AOy*axis_vect[0])**2
            )

        ABv = ABx*axis_vect[0] + ABy*axis_vect[1] + ABz*axis_vect[2]
        AOv = AOx*axis_vect[0] + AOy*axis_vect[1] + AOz*axis_vect[2]
        ABAO = ABx*AOx + ABy*AOy + ABz*AOz
        B = ABv*AOv - ABAO

    else:
        AOx, AOy, AOz = None, None, None
        ABvn2, AOvn2, B = None, None, None
        i0 = None

    return (
        pts_x, pts_y, pts_z,
        ABx, ABy, ABz,
        AOx, AOy, AOz,
        ABvn2, AOvn2, B,
        i0,
    )


def _tangency_radius(
    coll=None,
    key=None,
    key_cam=None,
    quantity=None,
    # limits
    segment=None,
    lim_to_segments=None,
    # tangency radius
    axis_pt=None,
    axis_vect=None,
):

    # --------------
    # check inputs
    # --------------

    (
        key, quantity, segment, lim_to_segments, axis_pt, axis_vect,
    ) = _tangency_radius_check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        quantity=quantity,
        segment=segment,
        lim_to_segments=lim_to_segments,
        axis_pt=axis_pt,
        axis_vect=axis_vect,
    )

    # -----------
    # compute
    # -----------

    # --------
    # prepare

    (
     pts_x, pts_y, pts_z,
     ABx, ABy, ABz,
     AOx, AOy, AOz,
     ABvn2, AOvn2, B,
     i0,
     ) = _tangency_radius_prepare(
         coll=coll,
         key=key,
         quantity=quantity,
         segment=segment,
         axis_pt=axis_pt,
         axis_vect=axis_vect,
    )

    # --------
    # radius

    if quantity == 'tangency radius':
        kk = np.zeros(ABx.shape, dtype=float)
        iok = ABvn2 > 0.
        kk[iok] = -B[iok] / ABvn2[iok]

        # ------------------------
        # lim_to_segments

        if lim_to_segments is True:
            i0 = kk < 0.
            i1 = kk > 1.

            kk[i0] = 0.
            kk[i1] = 1.

        out = np.sqrt(kk**2 * ABvn2 + 2*kk*B + AOvn2)

    elif quantity == 'length':
        out = np.sqrt(ABx**2 + ABy**2 + ABz**2)
        kk = None

    else:
        kdata = coll.dobj['rays'][key]['alpha']
        out = coll.ddata[kdata]['data']
        kk = None

    # -------
    # ref

    if segment is None:
        ref = coll.dobj['rays'][key]['ref']
    elif segment.size == 1:
        ref = coll.dobj['rays'][key]['ref'][1:]
        out = out[0]
        if quantity == 'tangency radius':
            kk = kk[0]
    else:
        ref = None

    # -------
    # return

    return out, kk, ref


def intersect_radius(
    coll=None,
    key=None,
    key_cam=None,
    axis_pt=None,
    axis_vect=None,
    axis_radius=None,
    segment=None,
    lim_to_segments=None,
    return_pts=None,
    return_itot=None,
):

    # --------------
    # check inputs
    # --------------

    (
     key, quantity, segment, lim_to_segments, axis_pt, axis_vect,
     ) = _tangency_radius_check(
         coll=coll,
         key=key,
         key_cam=key_cam,
         quantity='tangency radius',
         axis_pt=axis_pt,
         axis_vect=axis_vect,
         segment=segment,
         lim_to_segments=lim_to_segments,
    )

    # axis_radius
    axis_radius = ds._generic_check._check_var(
        axis_radius, 'axis_radius',
        types=(float, int),
        sign='> 0.',
    )

    # return_pts
    return_pts = ds._generic_check._check_var(
        return_pts, 'return_pts',
        types=bool,
        default=False,
    )

    if return_pts is True and lim_to_segments is False:
        msg = (
            "return_pts requires lim_to_segments = True"
            )
        raise Exception(msg)

    # return_itot
    return_itot = ds._generic_check._check_var(
        return_itot, 'return_itot',
        types=bool,
        default=False,
    )

    # -----------
    # compute
    # -----------

    # --------
    # prepare

    (
     pts_x, pts_y, pts_z,
     ABx, ABy, ABz,
     AOx, AOy, AOz,
     ABvn2, AOvn2, B,
     i0,
     ) = _tangency_radius_prepare(
         coll=coll,
         key=key,
         quantity=quantity,
         segment=segment,
         axis_pt=axis_pt,
         axis_vect=axis_vect,
         )

    # pre-select according to rad_min
    kmin = np.zeros(ABx.shape, dtype=float)

    iok = ABvn2 > 0.
    kmin[iok] = -B[iok] / ABvn2[iok]
    rad_min2 = kmin**2 * ABvn2 + 2*kmin*B + AOvn2
    rad_min2[np.abs(rad_min2) < 1e-9] = 0.
    rad_min = np.sqrt(rad_min2)

    # prepare solutions
    iin = (rad_min < axis_radius) & iok

    # there can be up to 2 solutions
    k0 = np.full(ABx.shape, np.nan)
    k1 = np.full(ABx.shape, np.nan)

    if np.any(iin):
        delta = B[iin]**2 - ABvn2[iin]*(AOvn2[iin] - axis_radius**2)

        k0[iin] = (-B[iin] - np.sqrt(delta)) / ABvn2[iin]
        k1[iin] = (-B[iin] + np.sqrt(delta)) / ABvn2[iin]

    # ----------------
    # lim_to_segments

    if lim_to_segments is True:

        ind = np.copy(iin)
        ind[ind] = (k0[ind] < 0) & (k1[ind] > 0)
        k0[ind] = 0.

        ind = np.copy(iin)
        ind[ind] = (k0[ind] < 1) & (k1[ind] > 1)
        k1[ind] = 1.

        ind = np.copy(iin)
        ind[ind] = (k0[ind] >= 0.) & (k0[ind] <= 1.)
        k0[~ind] = np.nan

        ind = np.copy(iin)
        ind[ind] = (k1[ind] >= 0.) & (k1[ind] <= 1.)
        k1[~ind] = np.nan

    # ----------------------
    # additional derivations

    iok = np.isfinite(k0) & np.isfinite(k1)

    if return_pts is True:

        iok2 = np.copy(iok)
        iok2[iok] = k1[iok] < 1.

        false = np.zeros(tuple(np.r_[1, iok.shape[1:]]), dtype=bool)
        iok0 = np.concatenate((iok, false), axis=0)
        iok02 = np.concatenate((iok2, false), axis=0)
        iok12 = np.concatenate((false, iok2), axis=0)

        # Make sure there a single continued sequence per ray 
        # build index and check continuity

        px = np.full(pts_x.shape, np.nan)
        py = np.full(pts_x.shape, np.nan)
        pz = np.full(pts_x.shape, np.nan)

        px[iok0] = pts_x[iok0] + k0[iok] * ABx[iok]
        py[iok0] = pts_y[iok0] + k0[iok] * ABy[iok]
        pz[iok0] = pts_z[iok0] + k0[iok] * ABz[iok]

        px[iok12] = pts_x[iok02] + k1[iok2] * ABx[iok2]
        py[iok12] = pts_y[iok02] + k1[iok2] * ABy[iok2]
        pz[iok12] = pts_z[iok02] + k1[iok2] * ABz[iok2]
        px[-1, ...] = pts_x[-2, ...] + k1[-1, ...] * ABx[-1, ...]
        py[-1, ...] = pts_y[-2, ...] + k1[-1, ...] * ABy[-1, ...]
        pz[-1, ...] = pts_z[-2, ...] + k1[-1, ...] * ABz[-1, ...]

    if return_itot is True:

        i02 = np.full(pts_x.shape, np.nan)
        for ii in range(pts_x.shape[0]):
            i02[ii, ...] = i0[ii]

        itot = np.full(pts_x.shape, np.nan)

        itot[iok0] = i02[iok0] + k0[iok]
        itot[iok12] = i02[iok02] + k1[iok2]
        itot[-1, ...] = i02[-2, ...] + k1[-1, ...]

    # ------
    # return

    if return_pts is True and return_itot is True:
        return k0, k1, iok, px, py, pz, itot
    elif return_pts is True:
        return k0, k1, iok, px, py, pz
    elif return_itot is True:
        return k0, k1, iok, itot
    else:
        return k0, k1, iok



# -*- coding: utf-8 -*-


# Built-in
# import copy


# Common
import numpy as np
import datastock as ds


# tofu
from ._class01_Plasma2D import Plasma2D as Previous
from . import _class2_check as _check
from . import _class2_compute as _compute
from . import _class2_plot as _plot


__all__ = ['Rays']


# #############################################################################
# #############################################################################
#                           Diagnostic
# #############################################################################


class Rays(Previous):

    _show_in_summary = 'all'

    _dshow = dict(Previous._dshow)
    _dshow.update({
        'rays': [
            'shape',
            'ref',
            'pts',
            'lamb',
            'Rmin',
            'alpha',
            'dalpha',
            'dbeta',
        ],
    })

    def add_rays(
        self,
        key=None,
        # start points
        start_x=None,
        start_y=None,
        start_z=None,
        # wavelength
        lamb=None,
        # ref
        ref=None,
        # from pts
        pts_x=None,
        pts_y=None,
        pts_z=None,
        # from ray-tracing (vect + length or config or diag)
        vect_x=None,
        vect_y=None,
        vect_z=None,
        length=None,
        config=None,
        reflections_nb=None,
        reflections_type=None,
        key_nseg=None,
        diag=None,
        key_cam=None,
    ):
        """ Add a set of rays

        Can be defined from:
            - pts: explicit
            - starting pts + vectors

        """

        # check / format input
        dref, ddata, dobj = _check._rays(
            coll=self,
            key=key,
            # start points
            start_x=start_x,
            start_y=start_y,
            start_z=start_z,
            # wavelength
            lamb=lamb,
            # ref
            ref=ref,
            # from pts
            pts_x=pts_x,
            pts_y=pts_y,
            pts_z=pts_z,
            # from ray-tracing (vect + length or config or diag)
            vect_x=vect_x,
            vect_y=vect_y,
            vect_z=vect_z,
            length=length,
            config=config,
            reflections_nb=reflections_nb,
            reflections_type=reflections_type,
            key_nseg=key_nseg,
            diag=diag,
            key_cam=key_cam,
        )

        # update dicts
        self.update(dref=dref, ddata=ddata, dobj=dobj)

    # -----------------
    # remove
    # -----------------

    def remove_rays(self, key=None):
        return _check._remove(
            coll=self,
            key=key,
        )

    # --------------
    #  utilities
    # --------------

    def get_rays_start(
        self,
        key=None,
        key_cam=None,
    ):
        return _check._get_start(coll=self, key=key, key_cam=key_cam)

    def get_rays_pts(
        self,
        key=None,
        key_cam=None,
    ):
        return _check._get_pts(coll=self, key=key, key_cam=key_cam)

    def get_rays_vect(
        self,
        key=None,
        key_cam=None,
        norm=None,
    ):
        return _check._get_vect(coll=self, key=key, key_cam=key_cam, norm=norm)

    # --------------
    # discretizing
    # --------------

    def sample_rays(
        self,
        key=None,
        key_cam=None,
        res=None,
        mode=None,
        segment=None,
        ind_flat=None,
        radius_max=None,
        concatenate=None,
        return_coords=None,
    ):
        """ Return the sampled rays

        Parameters
        ----------
        key:        str
            key of the rays / diag los to sample
        res:        float
            sampling resolution
        mode:       str
            sampling mode
                - 'rel': relative, res is in [0, 1], 0.1 = 10 samples / segment
                - 'abs': absolute, res is a distance in m
        segment:    None / int / iterable of ints
            indices of te segments to be sampled
                - None: all
                - int: a single segment
                - iterable of ints: several segments
            Typical usage: None or -1 (last segment)
        radius_max:     None / float
            If provided, only sample the portion of segments that are inside
            the provided ;ajor radius
        conctenate:     bool
            flag indicating whether to concatenate the sampled points per ray
        """

        return _compute._sample(
            coll=self,
            key=key,
            key_cam=key_cam,
            res=res,
            mode=mode,
            segment=segment,
            ind_flat=ind_flat,
            radius_max=radius_max,
            concatenate=concatenate,
            return_coords=return_coords,
        )

    # --------------
    # tangency radius
    # --------------

    def get_rays_quantity(
        self,
        key=None,
        key_cam=None,
        quantity=None,
        # limits
        segment=None,
        lim_to_segments=None,
        # for tangency radius
        axis_pt=None,
        axis_vect=None,
    ):
        """ Return a ray-specific quantity of each ray segment

        parameters
        ----------
        quantity:   str
            - 'length': length of rays
            - 'tangency radius': the tangency radius to an axis
        axis_pt:    len=3 iterable
            (x, y, z) coordinates of a pt on the axis, default to [0, 0, 0]
        axis_vect:  len=3 iterable
            (x, y, z) coordinates of the axis vector, default to [0, 0, 1]
        lim_to_segments: bool
            flag indicating whether to limit solutions to the segments

        Return
        -------
        radius:     np.ndarray of floats
            the tangency radii
        kk:         np.ndarray of floats
            the normalized longitudinal coordinate of the tangency points
        ref:        tuple
            The ref tuple on which the data depends
        """

        return _compute._tangency_radius(
            coll=self,
            key=key,
            key_cam=key_cam,
            quantity=quantity,
            # limits
            segment=segment,
            lim_to_segments=lim_to_segments,
            # for tangency radius
            axis_pt=axis_pt,
            axis_vect=axis_vect,
        )

    def get_rays_intersect_radius(
        self,
        key=None,
        key_cam=None,
        axis_pt=None,
        axis_vect=None,
        axis_radius=None,
        segment=None,
        lim_to_segments=None,
        return_pts=None,
        return_itot=None,
    ):
        """ Return the tangancy radius to an axis of each ray segment
        
        parameters
        ----------
        axis_pt:    len=3 iterable
            (x, y, z) coordinates of a pt on the axis, default to [0, 0, 0]
        axis_vect:  len=3 iterable
            (x, y, z) coordinates of the axis vector, default to [0, 0, 1]
        axis_radius:    float
            The radius around the axis defining the cylinder to intersect
        lim_to_segments: bool
            flag indicating whether to limit solutions to the segments
        return_pts:
            flag indicating whether to return the pts (x, y, z) coordinates 
                
        Return
        -------
        k0:         np.ndarray of floats
            First solution, per segment
        k1:         np.ndarray of floats
            Second solution, per segment
        iok:        np.ndarray of bool
            Flag indicating which segments have at least an intersection
        pts_x:      np.ndarray of floats
            The x coordinates of the points inside the cylinder
        pts_y:      np.ndarray of floats
            The y coordinates of the points inside the cylinder
        pts_z:      np.ndarray of floats
            The z coordinates of the points inside the cylinder

        """
        
        return _compute.intersect_radius(
            coll=self,
            key=key,
            key_cam=key_cam,
            axis_pt=axis_pt,
            axis_vect=axis_vect,
            axis_radius=axis_radius,
            segment=segment,
            lim_to_segments=lim_to_segments,
            return_pts=return_pts,
            return_itot=return_itot,
            )

    # --------------
    # plotting
    # --------------

    def plot_rays(
        self,
        key=None,
        proj=None,
        concatenate=None,
        mode=None,
        res=None,
        # config
        plot_config=None,
        # figure
        dax=None,
        dmargin=None,
        fs=None,
        wintit=None,
        # interactivity
        color_dict=None,
        nlos=None,
        dinc=None,
        connect=None,
    ):
        return _plot._plot_rays(
            coll=self,
            key=key,
            proj=proj,
            concatenate=concatenate,
            mode=mode,
            res=res,
            # config
            plot_config=plot_config,
            # figure
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            wintit=wintit,
            # interactivity
            color_dict=color_dict,
            nlos=nlos,
            dinc=dinc,
            connect=connect,
        )



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import scipy.interpolate as scpinterp


# specific
from . import _generic_check
from . import _utils_bsplines
from . import _class1_bsplines_operators_rect


# #############################################################################
# #############################################################################
#                       BivariateSplineRect - scipy subclass
# #############################################################################


class BivariateSplineRect(scpinterp.BivariateSpline):
    """ Subclass for tofu

    Defined from knots (unique) and deg
    coefs set to 1 by default

    Used self.set_coefs() to update
    """

    def __init__(self, knotsR=None, knotsZ=None, deg=None, shapebs=None):

        assert np.allclose(np.unique(knotsR), knotsR)
        assert np.allclose(np.unique(knotsZ), knotsZ)
        assert deg in [0, 1, 2, 3]

        # get knots pr bs
        self._get_knots_per_bs_for_basis_elements(
            knotsR=knotsR,
            knotsZ=knotsZ,
            deg=deg,
        )

        # full knots with multiplicity
        knotsR, nbsR = _utils_bsplines._get_knots_per_bs(
            knotsR, deg=deg, returnas='data', return_unique=True,
        )
        knotsZ, nbsZ = _utils_bsplines._get_knots_per_bs(
            knotsZ, deg=deg, returnas='data', return_unique=True,
        )

        coefs = np.ones((nbsR*nbsZ,), dtype=float)

        self.__nbs = (nbsR, nbsZ)
        self.tck = [knotsR, knotsZ, coefs]
        self.degrees = [deg, deg]

        # shapebs
        self.shapebs = shapebs

    def _get_knots_per_bs_for_basis_elements(
        self,
        knotsR=None,
        knotsZ=None,
        deg=None,
    ):

        # added for details
        knots_per_bs_x = _utils_bsplines._get_knots_per_bs(
            knotsR, deg=deg, returnas='data',
        )
        knots_per_bs_y = _utils_bsplines._get_knots_per_bs(
            knotsZ, deg=deg, returnas='data',
        )

        self.knots_per_bs_x = knots_per_bs_x
        self.knots_per_bs_y = knots_per_bs_y

        if deg == 0:
            pass
        else:
            knots_per_bs_x = np.concatenate(
                (
                    np.tile(knots_per_bs_x[0, :] - 1, (deg, 1)),
                    knots_per_bs_x,
                    np.tile(knots_per_bs_x[-1, :] + 1, (deg, 1)),
                ),
                axis=0,
            )
            knots_per_bs_y = np.concatenate(
                (
                    np.tile(knots_per_bs_y[0, :] - 1, (deg, 1)),
                    knots_per_bs_y,
                    np.tile(knots_per_bs_y[-1, :] + 1, (deg, 1)),
                ),
                axis=0,
            )

        self.knots_per_bs_x_pad = np.asfortranarray(knots_per_bs_x)
        self.knots_per_bs_y_pad = np.asfortranarray(knots_per_bs_y)

    def _check_coefs(self, coefs=None):
        """ None for ev_details, (nt, shapebs) for sum """
        if coefs is not None:
            assert coefs.ndim == len(self.shapebs) + 1
            assert coefs.shape[1:] == self.shapebs

    def set_coefs(
        self,
        coefs=None,
        cropbs_neg_flat=None,
    ):

        nbs = self.__nbs[0]*self.__nbs[1]
        if coefs.shape == self.shapebs:
            self.tck[2][...] = coefs.ravel()
        elif coefs.shape == (self.nbs,):
            self.tck[2][...] = coefs
        else:
            msg = f"Wrong coefs shape!\nProvided: {coefs.shape}"
            raise Exception(msg)

        # ------------
        # crop and set

        if cropbs_neg_flat is not None:
            self.tck[2][cropbs_neg_flat] = 0.

    def __call__(
        self,
        R=None,
        Z=None,
        coefs=None,
        crop=None,
        cropbs=None,
        val_out=None,
        # for compatibility (unused)
        indbs_tf=None,
    ):

        if val_out is None:
            val_out = np.nan

        # coefs
        self._check_coefs(coefs=coefs)

        # r, z
        r, z, crop = _check_RZ_crop(
            R=R,
            Z=Z,
            crop=crop,
            cropbs=cropbs,
        )

        # prepare
        nt = coefs.shape[0]
        shape = tuple(np.r_[nt, r.shape])
        val = np.zeros(shape, dtype=float)
        cropbs_neg_flat = ~cropbs.ravel() if crop else None

        # interpolate
        for ii in range(nt):
            # prepare
            self.set_coefs(
                coefs=coefs[ii, ...],
                cropbs_neg_flat=cropbs_neg_flat,
            )

            # compute
            val[ii, ...] = super().__call__(r, z, grid=False)

        # clean
        if val_out is not False:
            indout = (
                (r < self.tck[0][0]) | (r > self.tck[0][-1])
                | (z < self.tck[1][0]) | (z > self.tck[1][-1])
            )
            val[:, indout] = val_out
        return val

    def ev_details(
        self,
        R=None,
        Z=None,
        indbs_tf=None,
        crop=None,
        cropbs=None,
        # for compatibility (unused)
        coefs=None,
        val_out=None,
    ):
        """
        indbs_tf = (ar0, ar1)
            tuple of 2 flat arrays of int (for R and Z)
        """

        # -----------
        # check input

        x, y, crop = _check_RZ_crop(
            R=R,
            Z=Z,
            crop=crop,
            cropbs=cropbs,
        )

        # -----------
        # prepare

        deg = self.degrees[0]
        nbs = indbs_tf[0].size
        shape = x.shape
        x = np.ascontiguousarray(x.ravel(), dtype=float)
        y = np.ascontiguousarray(y.ravel(), dtype=float)
        coef = np.zeros((deg + 4, 1), dtype=float)
        coef[deg] = 1.
        outy = np.full((x.size, 1), np.nan)

        # -----------
        # compute

        val = np.zeros(tuple(np.r_[x.size, nbs]))
        indtot = np.arange(0, nbs)

        iz_u = np.unique(indbs_tf[1])

        for iz in iz_u:

            scpinterp._bspl.evaluate_spline(
                self.knots_per_bs_y_pad[:, iz],
                coef,
                self.degrees[1],
                y,
                0,
                False,
                outy,
            )

            indoky = ~np.isnan(outy)
            if not np.any(indoky):
                continue
            indokx = np.copy(indoky)

            indr = indbs_tf[1] == iz
            ir = indbs_tf[0][indr]
            for ii, iir in enumerate(ir):

                if ii > 0:
                    indokx[...] = indoky

                outx = np.full((indoky.sum(), 1), np.nan)

                scpinterp._bspl.evaluate_spline(
                    self.knots_per_bs_x_pad[:, iir],
                    coef,
                    self.degrees[0],
                    x[indoky[:, 0]],
                    0,
                    False,
                    outx,
                )

                ixok = ~np.isnan(outx)
                if not np.any(ixok):
                    continue

                indokx[indoky] = ixok[:, 0]
                val[indokx[:, 0], indtot[indr][ii]] = (outx[ixok]*outy[indokx])

        if shape != x.shape:
            val = np.reshape(val, tuple(np.r_[shape, -1]))

        return val

    def get_overlap(self):
        return _get_overlap(
            deg=self.degrees[0],
            knotsx=self.knots_per_bs_x,
            knotsy=self.knots_per_bs_y,
            shapebs=self.shapebs,
        )

    def get_operator(
        self,
        operator=None,
        geometry=None,
        cropbs_flat=None,
        # specific to deg = 0
        cropbs=None,
        centered=None,
        # to return gradR, gradZ, for D1N2 deg 0, for tomotok
        returnas_element=None,
    ):
        """ Get desired operator """
        return _class1_bsplines_operators_rect.get_mesh2dRect_operators(
            deg=self.degrees[0],
            operator=operator,
            geometry=geometry,
            knotsx_mult=self.tck[0],
            knotsy_mult=self.tck[1],
            knotsx_per_bs=self.knots_per_bs_x,
            knotsy_per_bs=self.knots_per_bs_y,
            overlap=self.get_overlap(),
            cropbs_flat=cropbs_flat,
            # specific to deg = 0
            cropbs=cropbs,
            centered=centered,
            # to return gradR, gradZ, for D1N2 deg 0, for tomotok
            returnas_element=returnas_element,
        )


# #############################################################################
# #############################################################################
#                       Mesh2DRect - bsplines - overlap
# #############################################################################


def _get_overlap(
    deg=None,
    knotsx=None,
    knotsy=None,
    shapebs=None,
):
    # nb of overlapping, inc. itself in 1d
    nbsR, nbsZ = shapebs
    indR0 = np.tile(np.arange(0, nbsR), nbsZ)
    indZ0 = np.repeat(np.arange(0, nbsZ), nbsR)

    # complete
    ntot = 2*deg + 1

    addR = np.tile(np.arange(-deg, deg+1), ntot)
    addZ = np.repeat(np.arange(-deg, deg+1), ntot)

    interR = indR0[None, :] + addR[:, None]
    interZ = indZ0[None, :] + addZ[:, None]

    # purge
    inter = interR + interZ*nbsR
    indneg = (
        (interR < 0) | (interR >= nbsR) | (interZ < 0) | (interZ >= nbsZ)
    )
    inter[indneg] = -1

    return inter


# #############################################################################
# #############################################################################
#                           Mesh2DRect - bsplines
# #############################################################################


def _check_RZ_crop(
    R=None,
    Z=None,
    crop=None,
    cropbs=None,
):

    # R, Z
    if not isinstance(R, np.ndarray):
        R = np.atleast_1d(R)
    if not isinstance(Z, np.ndarray):
        Z = np.atleast_1d(Z)
    assert R.shape == Z.shape

    # crop
    if crop is None:
        crop = True
    if not isinstance(crop, bool):
        msg = (
            "Arg crop must be a bool!\n"
            f"Provided: {crop}"
        )
        raise Exception(msg)
    crop = crop and cropbs is not None and cropbs is not False

    return R, Z, crop


def get_bs2d_RZ(deg=None, Rknots=None, Zknots=None):

    # ----------------
    # get knots per bspline, nb of bsplines...

    knots_per_bs_R = _utils_bsplines._get_knots_per_bs(
        Rknots, deg=deg, returnas='data',
    )
    knots_per_bs_Z = _utils_bsplines._get_knots_per_bs(
        Zknots, deg=deg, returnas='data',
    )
    nbkbs = knots_per_bs_R.shape[0]
    shapebs = (knots_per_bs_R.shape[1], knots_per_bs_Z.shape[1])

    # ----------------
    # get centers of bsplines

    Rbs_apex = _utils_bsplines._get_apex_per_bs(
        knots=Rknots,
        knots_per_bs=knots_per_bs_R,
        deg=deg
    )
    Zbs_apex = _utils_bsplines._get_apex_per_bs(
        knots=Zknots,
        knots_per_bs=knots_per_bs_Z,
        deg=deg
    )
    return shapebs, Rbs_apex, Zbs_apex, knots_per_bs_R, knots_per_bs_Z


def get_bs2d_func(
    deg=None,
    Rknots=None,
    Zknots=None,
    shapebs=None,
    # knots_per_bs_R=None,
    # knots_per_bs_Z=None,
):

    # ----------------
    # Define functions

    clas = BivariateSplineRect(
        knotsR=Rknots,
        knotsZ=Zknots,
        deg=deg,
        shapebs=shapebs,
    )

    return clas.ev_details, clas.__call__, clas




"""
This module contains tests for tofu.geom in its structured version
"""

# Built-in
import os
import warnings
import itertools as itt

# Standard
import numpy as np
import scipy.constants as scpct
import matplotlib.pyplot as plt

# tofu-specific
from tofu import __version__
import tofu.spectro as tfs

_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.spectro'


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    print("")   # this is to get a newline after the dots
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if not lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following previous test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    # print("setup_module before anything in this file")


def teardown_module(module):
    # os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    # os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    # print("teardown_module after everything in this file")
    # print("") # this is to get a newline
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    pass


# def my_setup_function():
#    print ("my_setup_function")

# def my_teardown_function():
#    print ("my_teardown_function")

# @with_setup_method(my_setup_function, my_teardown_function)
# def test_numbers_3_4():
#    print 'test_numbers_3_4  <============================ actual test code'
#    assert multiply(3,4) == 12

# @with_setup_method(my_setup_function, my_teardown_function)
# def test_strings_a_3():
#    print 'test_strings_a_3  <============================ actual test code'
#    assert multiply('a',3) == 'aaa'


#######################################################
#
#     Creating Ves objects and testing methods
#
#######################################################


class Test01_RockingCurve(object):

    @classmethod
    def setup_class(cls):

        cls.lc = [
            k0 for k0 in tfs._rockingcurve_def._DCRYST.keys()
            if 'xxx' not in k0.lower()
        ]

    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_rockingcurve_compute(self):

        for k0 in self.lc:
            dout = tfs.compute_rockingcurve(
                crystal=k0,
                lamb=np.r_[3.969067],
                miscut=False,
                alpha_limits=None,
                therm_exp=False,
                temp_limits=None,
                # Plot
                plot_therm_exp=False,
                plot_asf=False,
                plot_power_ratio=False,
                plot_asymmetry=False,
                plot_cmaps=False,
                # return
                returnas=dict,
            )
